"""
TODO:
 - AsyncExitStacked
"""
import abc
import contextlib
import contextvars
import functools
import threading
import time
import types
import typing as ta


T = ta.TypeVar('T')


##


class _NOT_SET:  # noqa
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError


class ContextManaged:
    def __enter__(self) -> ta.Self:
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> bool | None:
        return None


class NopContextManager(ContextManaged):
    def __init__(self, /, value: ta.Any = _NOT_SET) -> None:
        super().__init__()

        self._value = value

    def __enter__(self):
        if (value := self._value) is _NOT_SET:
            return self
        else:
            return value

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        raise TypeError


NOP_CONTEXT_MANAGER = NopContextManager()


##


class ContextManager(abc.ABC, ta.Generic[T]):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if not hasattr(cls.__contextmanager__, '_is_contextmanager'):
            cls.__contextmanager__ = contextlib.contextmanager(cls.__contextmanager__)  # type: ignore  # noqa
            cls.__contextmanager__._is_contextmanager = True  # type: ignore  # noqa

    @abc.abstractmethod
    def __contextmanager__(self) -> ta.Iterable[T]:
        raise NotImplementedError

    __contextmanager__._is_contextmanager = True  # type: ignore  # noqa

    _contextmanager: ta.Any

    def __enter__(self) -> T:
        self._contextmanager = self.__contextmanager__()
        return self._contextmanager.__enter__()  # type: ignore

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> bool | None:
        return self._contextmanager.__exit__(exc_type, exc_val, exc_tb)


class AsyncContextManager(abc.ABC, ta.Generic[T]):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if not hasattr(cls.__asynccontextmanager__, '_is_asynccontextmanager'):
            cls.__asynccontextmanager__ = contextlib.asynccontextmanager(cls.__asynccontextmanager__)  # type: ignore  # noqa
            cls.__asynccontextmanager__._is_asynccontextmanager = True  # type: ignore  # noqa

    @abc.abstractmethod
    def __asynccontextmanager__(self) -> ta.AsyncIterator[T]:
        raise NotImplementedError

    __asynccontextmanager__._is_asynccontextmanager = True  # type: ignore  # noqa

    _asynccontextmanager: ta.Any

    async def __aenter__(self) -> T:
        self._asynccontextmanager = self.__asynccontextmanager__()
        return await self._asynccontextmanager.__aenter__()  # type: ignore

    async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> None:
        return await self._asynccontextmanager.__aexit__(exc_type, exc_val, exc_tb)


##


@contextlib.contextmanager
def maybe_managing(obj: T) -> ta.Iterator[T]:
    if isinstance(obj, ta.ContextManager):
        with obj:
            yield ta.cast(T, obj)
    else:
        yield obj


@contextlib.contextmanager
def disposing(obj: T, attr: str = 'dispose') -> ta.Iterator[T]:
    try:
        yield obj
    finally:
        getattr(obj, attr)()


@contextlib.contextmanager
def breakpoint_on_exception():  # noqa
    try:
        yield
    except Exception as e:  # noqa
        breakpoint()  # noqa
        raise


@contextlib.contextmanager
def context_var_setting(var: contextvars.ContextVar[T], val: T) -> ta.Iterator[T]:
    token = var.set(val)
    try:
        yield val
    finally:
        var.reset(token)


##


ContextWrappable: ta.TypeAlias = ta.ContextManager | str | ta.Callable[..., ta.ContextManager]


class ContextWrapped:
    def __init__(self, fn: ta.Callable, cm: str | ContextWrappable) -> None:
        super().__init__()

        self._fn = (fn,)
        self._cm = cm
        self._name: str | None = None

        functools.update_wrapper(self, fn)

    @property
    def _fn(self):
        return self.__fn

    @_fn.setter
    def _fn(self, x):
        self.__fn = x

    def __set_name__(self, owner, name):
        if name is not None:
            if self._name is not None:
                if name != self._name:
                    raise NameError(name, self._name)
            else:
                self._name = name

    def __get__(self, instance, owner=None):
        if instance is None and owner is None:
            return self
        fn = self._fn[0].__get__(instance, owner)  # noqa
        cm: ta.Any = self._cm
        if isinstance(self._cm, str):
            if instance is not None:
                cm = getattr(instance, cm)
            elif owner is not None:
                cm = getattr(owner, cm)
            else:
                raise TypeError(cm)
        elif hasattr(cm, '__enter__'):
            pass
        elif callable(cm):
            cm = cm.__get__(instance, owner)  # noqa
        else:
            raise TypeError(cm)
        ret = type(self)(fn, cm)
        if self._name is not None:
            with contextlib.suppress(TypeError):
                instance.__dict__[self._name] = ret
        return ret

    def __call__(self, *args, **kwargs):
        if isinstance(self._cm, str):
            raise TypeError(self._cm)
        cm = self._cm
        if not hasattr(cm, '__enter__') and callable(cm):
            cm = cm(*args, **kwargs)
        with cm:  # type: ignore
            return self._fn[0](*args, **kwargs)


def context_wrapped(cm):  # ContextWrappable -> ta.Callable[[CallableT], CallableT]:  # noqa
    def inner(fn):
        return ContextWrapped(fn, cm)
    return inner


##


Lockable = ta.Callable[[], ta.ContextManager]
DefaultLockable = bool | Lockable | ta.ContextManager | None


def default_lock(value: DefaultLockable, default: DefaultLockable = None) -> Lockable:
    if value is None:
        value = default

    if value is True:
        lock = threading.RLock()
        return lambda: lock

    elif value is False or value is None:
        return lambda: NOP_CONTEXT_MANAGER

    elif callable(value):
        return value

    elif isinstance(value, ta.ContextManager):
        return lambda: value

    else:
        raise TypeError(value)


##


class Timer:
    def __init__(
            self,
            clock: ta.Callable[[], float] | None = None,
    ) -> None:
        super().__init__()
        self._clock = clock if clock is not None else time.monotonic

    _start: float
    _end: float

    @property
    def start(self) -> float:
        return self._start

    @property
    def end(self) -> float:
        return self._end

    @property
    def elapsed(self) -> float:
        return self._end - self._start

    def __enter__(self) -> ta.Self:
        self._start = self._clock()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._end = self._clock()



import collections.abc
import dataclasses as dc
import typing as ta


##


@dc.dataclass(frozen=True)
class QualifiedName(ta.Sequence[str]):
    parts: ta.Sequence[str]

    def __post_init__(self) -> None:
        if not (
                self.parts and
                not isinstance(self.parts, str) and
                all(self.parts) and
                all(isinstance(p, str) for p in self.parts)
        ):
            raise ValueError(self)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}([{", ".join(map(repr, self.parts))}])'

    @property
    def dotted(self) -> str:
        return '.'.join(self.parts)

    def prefixed(self, sz: int) -> tuple[str | None, ...]:
        if len(self) > sz:
            raise ValueError(self)
        return ((None,) * (sz - len(self))) + tuple(self.parts)

    @property
    def pair(self) -> tuple[str | None, str]:
        return self.prefixed(2)  # type: ignore

    @property
    def triple(self) -> tuple[str | None, str | None, str]:
        return self.prefixed(3)  # type: ignore

    @property
    def quad(self) -> tuple[str | None, str | None, str | None, str]:
        return self.prefixed(4)  # type: ignore

    def __iter__(self) -> ta.Iterator[str]:
        return iter(self.parts)

    def __len__(self) -> int:
        return len(self.parts)

    def __getitem__(self, idx: int) -> str:  # type: ignore
        return self.parts[idx]

    @classmethod
    def of_dotted(cls, dotted: str) -> 'QualifiedName':
        return cls(dotted.split('.'))

    @classmethod
    def of(
            cls,
            obj: ta.Union['QualifiedName', ta.Iterable[str]],
    ) -> 'QualifiedName':
        if isinstance(obj, QualifiedName):
            return obj
        elif isinstance(obj, str):
            raise TypeError(obj)
        elif isinstance(obj, collections.abc.Iterable):
            return cls(list(obj))
        else:
            raise TypeError(obj)

    @classmethod
    def of_optional(
            cls,
            obj: ta.Union['QualifiedName', ta.Iterable[str], None],
    ) -> ta.Optional['QualifiedName']:
        if obj is None:
            return None
        else:
            return cls.of(obj)


def qn(*args: str) -> QualifiedName:
    return QualifiedName(args)



from ....diag import pydevd as opd
from ._registry import register


@register
class PydevdPlugin:
    def pytest_addoption(self, parser):
        parser.addoption('--no-pydevd', action='store_true', default=False, help='Disables pydevd debugging')

    def pytest_collection(self, session):
        if opd.get_setup() is not None:
            if hasattr(session.config, '_env_timeout'):
                session.config._env_timeout = None  # noqa

            # if (dbg := opd.get_global_debugger()) is not None:
            #     dbg.set_unit_tests_debugging_mode()

    # def pytest_exception_interact(self, node, call, report):
    #     if opd.get_setup() is not None:
    #         if not node.session.config.option.no_pydevd:
    #             opd.debug_unhandled_exception(call.excinfo._excinfo)  # noqa
    #
    #     return report



#!/usr/bin/env python3
"""
Some environments refuse to support running modules rather than scripts, so this is just "python -m" functionality
exposed as a script.
"""
import runpy
import sys


def _main() -> int:
    # Run the module specified as the next command line argument
    if len(sys.argv) < 2:
        print('No module specified for execution', file=sys.stderr)
        return 1

    if sys.argv[1] == '--wait':
        import os
        print(os.getpid())
        input()
        sys.argv.pop(1)

    del sys.argv[0]  # Make the requested module sys.argv[0]
    runpy._run_module_as_main(sys.argv[0])  # type: ignore  # noqa
    return 0


if __name__ == '__main__':
    sys.exit(_main())



import collections

from .. import lang


##


Tag = collections.namedtuple('Tag', 'tag')  # noqa


class Scope(lang.Abstract):
    def __repr__(self) -> str:
        return type(self).__name__


class Unscoped(Scope, lang.Singleton, lang.Final):
    pass



import typing as ta

from .base import BooleanKeyword
from .base import Keyword
from .base import Keywords
from .base import KeywordsKeyword
from .base import NumberKeyword
from .base import StrKeyword
from .base import StrOrStrsKeyword
from .base import StrToKeywordsKeyword
from .unknown import UnknownKeyword


##


def render_keyword(kw: Keyword) -> dict[str, ta.Any]:
    if isinstance(kw, BooleanKeyword):
        return {kw.tag: kw.b}

    elif isinstance(kw, NumberKeyword):
        return {kw.tag: kw.n}

    elif isinstance(kw, StrKeyword):
        return {kw.tag: kw.s}

    elif isinstance(kw, StrOrStrsKeyword):
        if isinstance(kw.ss, str):
            return {kw.tag: kw.ss}
        else:
            return {kw.tag: list(kw.ss)}

    elif isinstance(kw, KeywordsKeyword):
        return {kw.tag: render_keywords(kw.kw)}

    elif isinstance(kw, StrToKeywordsKeyword):
        return {kw.tag: {k: render_keywords(v) for k, v in kw.m.items()}}

    elif isinstance(kw, UnknownKeyword):
        return {kw.tag: kw.value}

    else:
        raise TypeError(kw)


def render_keywords(kws: Keywords) -> dict[str, ta.Any]:
    dct: dict[str, ta.Any] = {}
    for kw in kws.lst:
        kwd = render_keyword(kw)
        [(tag, val)] = kwd.items()
        if tag in dct:
            raise KeyError(tag)
        dct[tag] = val
    return dct



import abc
import typing as ta


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class PersistentMap(ta.Generic[K, V], abc.ABC):
    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def __contains__(self, item: K) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, item: K) -> V:
        raise NotImplementedError

    @abc.abstractmethod
    def __iter__(self) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError

    @abc.abstractmethod
    def with_(self, k: K, v: V) -> 'PersistentMap[K, V]':
        raise NotImplementedError

    @abc.abstractmethod
    def without(self, k: K) -> 'PersistentMap[K, V]':
        raise NotImplementedError

    @abc.abstractmethod
    def default(self, k: K, v: V) -> 'PersistentMap[K, V]':
        raise NotImplementedError



"""
TODO:
 - Configurable? class Config crtc?
  - all config on specs? prob not desirable purity, prob want ~some~ global config if only defaults
   - ProcessorConfigSpecProviderFactoryThingy?
  - FIXME: for now these are processor ctor kwargs that are never overridden
"""
import typing as ta

from ... import check
from ... import lang
from ..specs import ClassSpec


T = ta.TypeVar('T')


##


ProcessingContextItemFactory: ta.TypeAlias = ta.Callable[['ProcessingContext'], ta.Any]


class ProcessingContext:
    def __init__(
            self,
            cls: type,
            cs: ClassSpec,
            item_factories: ta.Mapping[type, ProcessingContextItemFactory],
            *,
            options: ta.Sequence[ta.Any] | None = None,
    ) -> None:
        super().__init__()

        self._cls = cls
        self._cs = cs
        self._item_factories = item_factories

        options_dct: dict = {}
        for o in options or ():
            check.not_in(type(o), options_dct)
            options_dct[type(o)] = o
        self._options_dct = options_dct

        self._items: dict = {}

    @property
    def cls(self) -> type:
        return self._cls

    @property
    def cs(self) -> ClassSpec:
        return self._cs

    def __getitem__(self, ty: type[T]) -> T:
        try:
            return self._items[ty]
        except KeyError:
            pass

        fac = self._item_factories[ty]
        ret = fac(self)
        self._items[ty] = ret
        return ret

    def option(self, ty: type[T]) -> T | None:
        return self._options_dct.get(ty)


##


class Processor(lang.Abstract):
    def __init__(self, ctx: ProcessingContext) -> None:
        super().__init__()

        self._ctx = ctx

    def check(self) -> None:
        pass

    def process(self, cls: type) -> type:
        return cls



import typing as ta

from .. import concerns as _concerns  # noqa  # imported for registration
from ..generation import processor as gp
from ..specs import ClassSpec
from .base import ProcessingContext
from .base import Processor
from .registry import all_processing_context_item_factories
from .registry import ordered_processor_types


##


def drive_cls_processing(
        cls: type,
        cs: ClassSpec,
        *,
        plan_only: bool = False,
        verbose: bool = False,
) -> type:
    options: list[ta.Any] = []
    if plan_only:
        options.append(gp.PlanOnly(True))
    if verbose:
        options.append(gp.Verbose(True))

    #

    ctx = ProcessingContext(
        cls,
        cs,
        all_processing_context_item_factories(),
        options=options,
    )

    processors: list[Processor] = [
        proc_cls(ctx)
        for proc_cls in ordered_processor_types()
    ]

    for p in processors:
        p.check()

    ret = cls
    for p in processors:
        ret = p.process(ret)

    return ret



from . import consts  # noqa

from .clients import (  # noqa
    BaseHttpResponse,
    HttpClient,
    HttpClientError,
    HttpRequest,
    HttpResponse,
    HttpxHttpClient,
    StreamHttpResponse,
    UrllibHttpClient,
    client,
    close_response,
    closing_response,
    read_response,
    request,
)

from .cookies import (  # noqa
    CookieTooBigError,
    dump_cookie,
    parse_cookie,
)

from .dates import (  # noqa
    http_date,
    parse_date,
)

from .encodings import (  # noqa
    latin1_decode,
    latin1_encode,
)

from .headers import (  # noqa
    CanHttpHeaders,
    HttpHeaders,
    headers,
)

from .json import (  # noqa
    JSON_TAGGER,
    JsonTag,
    JsonTagger,
    json_dumps,
    json_loads,
)

from .multipart import (  # noqa
    MultipartData,
    MultipartEncoder,
    MultipartField,
)



import ast
import collections
import functools
import types
import typing as ta

from .. import check


T = ta.TypeVar('T')

K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


def repr_round_trip_value(v: T) -> T:
    r = repr(v)
    v2 = ast.literal_eval(r)
    if v != v2:
        raise ValueError(v)
    return v2


##


def set_qualname(cls: type, value: T) -> T:
    if isinstance(value, types.FunctionType):
        value.__qualname__ = f'{cls.__qualname__}.{value.__name__}'
    return value


def set_new_attribute(cls: type, name: str, value: ta.Any) -> bool:
    if name in cls.__dict__:
        return True
    set_qualname(cls, value)
    setattr(cls, name, value)
    return False


##


class SealableRegistry(ta.Generic[K, V]):
    def __init__(self) -> None:
        super().__init__()

        self._dct: dict[K, V] = {}
        self._sealed = False

    def seal(self) -> None:
        self._sealed = True

    def __setitem__(self, k: K, v: V) -> None:
        check.state(not self._sealed)
        check.not_in(k, self._dct)
        self._dct[k] = v

    def __getitem__(self, k: K) -> V:
        self.seal()
        return self._dct[k]

    def items(self) -> ta.Iterator[tuple[K, V]]:
        self.seal()
        return iter(self._dct.items())


##


def class_decorator(fn):
    @functools.wraps(fn)
    def inner(cls=None, *args, **kwargs):
        if cls is None:
            return lambda cls: fn(cls, *args, _kwargs=kwargs, **kwargs)  # noqa
        return fn(cls, *args, **kwargs)
    return inner


##


_EMPTY_MAPPING_PROXY: ta.Mapping = types.MappingProxyType({})


def chain_mapping_proxy(*ms: ta.Mapping) -> types.MappingProxyType:
    m: ta.Any
    if len(ms) > 1:
        m = collections.ChainMap(*ms)  # type: ignore[arg-type]
    elif ms:
        [m] = ms
    else:
        m = _EMPTY_MAPPING_PROXY

    return types.MappingProxyType(m)



import contextlib
import typing as ta


P = ta.ParamSpec('P')
T = ta.TypeVar('T')


##


async def call_with_async_exit_stack(  # type: ignore[return]  # ???
        fn: ta.Callable[ta.Concatenate[contextlib.AsyncExitStack, P], ta.Awaitable[T]],
        *args: ta.Any,
        **kwargs: ta.Any,
) -> T:
    async with contextlib.AsyncExitStack() as aes:
        return await fn(aes, *args, **kwargs)



import abc
import functools
import typing as ta

from .maybes import Maybe
from .maybes import empty
from .maybes import just


T = ta.TypeVar('T')
I = ta.TypeVar('I')
O = ta.TypeVar('O')
R = ta.TypeVar('R')
I_contra = ta.TypeVar('I_contra', contravariant=True)
O_co = ta.TypeVar('O_co', covariant=True)
R_co = ta.TypeVar('R_co', covariant=True)


##


def nextgen(g: T) -> T:
    next(g)  # type: ignore
    return g


def autostart(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        g = fn(*args, **kwargs)
        if (o := next(g)) is not None:
            raise TypeError(o)
        return g
    return inner


##


@ta.runtime_checkable
class GeneratorLike(ta.Protocol[O_co, I_contra, R_co]):
    def send(self, i: I_contra) -> O_co:  # Raises[StopIteration[R_co]]
        ...

    def close(self) -> None:
        ...


class GeneratorLike_(abc.ABC, ta.Generic[O, I, R]):  # noqa
    @abc.abstractmethod
    def send(self, i: I) -> O:  # Raises[StopIteration[R]]
        raise NotImplementedError

    def close(self) -> None:
        pass


@ta.overload
def adapt_generator_like(gl: GeneratorLike_[O, I, R]) -> ta.Generator[O, I, R]:
    ...


@ta.overload
def adapt_generator_like(gl: GeneratorLike[O, I, R]) -> ta.Generator[O, I, R]:
    ...


def adapt_generator_like(gl):
    try:
        i = yield
        while True:
            i = yield gl.send(i)
    except StopIteration as e:
        return e.value
    finally:
        gl.close()


##


class Generator(ta.Generator[O, I, R]):
    def __init__(self, g: ta.Generator[O, I, R]) -> None:
        super().__init__()
        self._g = g

    @property
    def g(self) -> ta.Generator[O, I, R]:
        return self._g

    value: R

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self._g)
        except StopIteration as e:
            self.value = e.value
            raise

    def send(self, v):
        try:
            return self._g.send(v)
        except StopIteration as e:
            self.value = e.value
            raise

    def throw(self, *args):
        try:
            return self._g.throw(*args)
        except StopIteration as e:
            self.value = e.value
            raise

    def close(self):
        self._g.close()


##


class CoroutineGenerator(ta.Generic[O, I, R]):
    def __init__(self, g: ta.Generator[O, I, R]) -> None:
        super().__init__()
        self._g = g

    @property
    def g(self) -> ta.Generator[O, I, R]:
        return self._g

    #

    def close(self) -> None:
        self._g.close()

    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._g.close()

    #

    class Output(ta.NamedTuple, ta.Generic[T]):
        v: T

        @property
        def is_return(self) -> bool:
            raise NotImplementedError

    class Yield(Output[T]):
        @property
        def is_return(self) -> bool:
            return False

    class Return(Output[T]):
        @property
        def is_return(self) -> bool:
            return True

    class Nothing:
        def __new__(cls):
            raise TypeError

    #

    def send(self, /, v: I | type[Nothing] = Nothing) -> Yield[O] | Return[R]:
        try:
            if v is self.Nothing:
                o = next(self._g)
            else:
                o = self._g.send(v)  # type: ignore[arg-type]
        except StopIteration as e:
            return self.Return(e.value)
        else:
            return self.Yield(o)

    def send_opt(self, v: I | None) -> Yield[O] | Return[R]:
        return self.send(v if v is not None else self.Nothing)

    def send_maybe(self, v: Maybe[I]) -> Yield[O] | Return[R]:
        return self.send(v.or_else(self.Nothing))

    def throw(self, v: BaseException) -> Yield[O] | Return[R]:
        try:
            o = self._g.throw(v)
        except StopIteration as e:
            return self.Return(e.value)
        else:
            return self.Yield(o)


corogen = CoroutineGenerator


##


class GeneratorMappedIterator(ta.Generic[O, I, R]):
    """
    Like a `map` iterator but takes a generator instead of a function. Provided generator *must* yield outputs 1:1 with
    inputs.

    Generator return value will be captured on `value` property - present if the generator stopped, absent if the
    iterator stopped.
    """

    def __init__(self, g: ta.Generator[O, I, R], it: ta.Iterator[I]) -> None:
        super().__init__()

        self._g = g
        self._it = it
        self._value: Maybe[R] = empty()

    @property
    def g(self) -> ta.Generator[O, I, R]:
        return self._g

    @property
    def it(self) -> ta.Iterator[I]:
        return self._it

    @property
    def value(self) -> Maybe[R]:
        return self._value

    def __iter__(self) -> ta.Iterator[O]:
        return self

    def __next__(self) -> O:
        i = next(self._it)
        try:
            o = self._g.send(i)
        except StopIteration as e:
            self._value = just(e.value)
            raise StopIteration from e
        return o


def genmap(g: ta.Generator[O, I, R], it: ta.Iterable[I]) -> GeneratorMappedIterator[O, I, R]:
    return GeneratorMappedIterator(g, iter(it))



"""
TODO:
 - check validity
"""
import dataclasses as dc
import enum
import inspect
import typing as ta

from .classes.abstract import Abstract
from .classes.restrict import Final
from .classes.restrict import Sealed
from .maybes import Maybe
from .maybes import empty
from .maybes import just


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True, unsafe_hash=True)
class Param(Sealed, Abstract):
    name: str

    annotation: Maybe = empty()

    prefix: ta.ClassVar[str] = ''

    @property
    def name_with_prefix(self) -> str:
        return f'{self.prefix}{self.name}'


#


@dc.dataclass(frozen=True, unsafe_hash=True)
class VarParam(Param, Abstract):
    pass


@dc.dataclass(frozen=True, unsafe_hash=True)
class ArgsParam(VarParam, Final):
    prefix: ta.ClassVar[str] = '*'


@dc.dataclass(frozen=True, unsafe_hash=True)
class KwargsParam(VarParam, Final):
    prefix: ta.ClassVar[str] = '**'


#


@dc.dataclass(frozen=True, unsafe_hash=True)
class ValParam(Param):
    default: Maybe = empty()


@dc.dataclass(frozen=True, unsafe_hash=True)
class PosOnlyParam(ValParam, Final):
    pass


@dc.dataclass(frozen=True, unsafe_hash=True)
class KwOnlyParam(ValParam, Final):
    pass


#


class ParamSeparator(enum.Enum):
    POS_ONLY = '/'
    KW_ONLY = '*'


##


def _inspect_empty_to_maybe(o: T) -> Maybe[T]:
    if o is inspect.Parameter.empty:
        return empty()
    else:
        return just(o)


class ParamSpec(ta.Sequence[Param], Final):
    def __init__(self, *ps: Param) -> None:
        super().__init__()

        self._ps = ps

        self._hash: int | None = None

        self._has_annotations: bool | None = None
        self._has_defaults: bool | None = None

        self._with_seps: tuple[Param | ParamSeparator, ...] | None = None

    #

    @classmethod
    def of_signature(
            cls,
            sig: inspect.Signature,
            *,
            offset: int = 0,
            strip_annotations: bool = False,
            strip_defaults: bool = False,
    ) -> 'ParamSpec':
        ps: list[Param] = []

        ip: inspect.Parameter
        for i, ip in enumerate(sig.parameters.values()):
            if i < offset:
                continue

            ann = _inspect_empty_to_maybe(ip.annotation) if not strip_annotations else empty()
            dfl = _inspect_empty_to_maybe(ip.default) if not strip_defaults else empty()

            if ip.kind == inspect.Parameter.POSITIONAL_ONLY:
                ps.append(PosOnlyParam(ip.name, ann, dfl))

            elif ip.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                ps.append(ValParam(ip.name, ann, dfl))

            elif ip.kind == inspect.Parameter.VAR_POSITIONAL:
                ps.append(ArgsParam(ip.name, ann))

            elif ip.kind == inspect.Parameter.KEYWORD_ONLY:
                ps.append(KwOnlyParam(ip.name, ann, dfl))

            elif ip.kind == inspect.Parameter.VAR_KEYWORD:
                ps.append(KwargsParam(ip.name, ann))

            else:
                raise ValueError(ip.kind)

        return cls(*ps)

    @classmethod
    def inspect(
            cls,
            obj: ta.Any,
            **kwargs: ta.Any,
    ) -> 'ParamSpec':
        return cls.of_signature(
            inspect.signature(obj),
            **kwargs,
        )

    #

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(map(repr, self._ps))})'

    def __hash__(self) -> int:
        if (h := self._hash) is not None:
            return h
        self._hash = h = hash(self._ps)
        return h

    def __eq__(self, other: object) -> bool:
        if type(other) is not ParamSpec:
            raise TypeError(other)
        return self._ps == other._ps  # noqa

    #

    @property
    def has_annotations(self) -> bool:
        if (ha := self._has_annotations) is not None:
            return ha
        self._has_annotations = ha = any(
            isinstance(p, (VarParam, ValParam)) and p.annotation.present
            for p in self._ps
        )
        return ha

    @property
    def has_defaults(self) -> bool:
        if (hd := self._has_defaults) is not None:
            return hd
        self._has_defaults = hd = any(
            isinstance(p, ValParam) and p.default.present
            for p in self._ps
        )
        return hd

    #

    @property
    def with_seps(self) -> ta.Sequence[Param | ParamSeparator]:
        if (ws := self._with_seps) is not None:
            return ws

        l: list[Param | ParamSeparator] = []
        needs_pos_only = False
        seen_kw_only = False
        for p in self._ps:
            if isinstance(p, PosOnlyParam):
                needs_pos_only = True
            elif needs_pos_only:
                l.append(ParamSeparator.POS_ONLY)
                needs_pos_only = False

            if isinstance(p, KwOnlyParam) and not seen_kw_only:
                l.append(ParamSeparator.KW_ONLY)
                seen_kw_only = True
            elif isinstance(p, KwargsParam):
                seen_kw_only = True

            l.append(p)

        self._with_seps = ws = tuple(l)
        return ws

    #

    @ta.overload
    def __getitem__(self, index: int) -> Param: ...

    @ta.overload
    def __getitem__(self, index: slice) -> ta.Sequence[Param]: ...

    def __getitem__(self, index):
        return self._ps[index]

    def __len__(self) -> int:
        return len(self._ps)


##


def param_render(
        p: Param | ParamSeparator,
        *,
        render_annotation: ta.Callable[[ta.Any], str] | None = None,
        render_default: ta.Callable[[ta.Any], str] | None = None,
) -> str:
    if isinstance(p, Param):
        ann_s: str | None = None
        if p.annotation.present:
            if render_annotation is None:
                raise ValueError(f'Param {p.name} has an annotation but no annotation renderer provided')
            ann_s = render_annotation(p.annotation.must())

        dfl_s: str | None = None
        if isinstance(p, ValParam) and p.default.present:
            if render_default is None:
                raise ValueError(f'Param {p.name} has a default but no default renderer provided')
            dfl_s = render_default(p.default.must())

        if ann_s is not None:
            if dfl_s is not None:
                return f'{p.name_with_prefix}: {ann_s} = {dfl_s}'
            else:
                return f'{p.name_with_prefix}: {ann_s}'
        elif dfl_s is not None:
            return f'{p.name_with_prefix}={dfl_s}'
        else:
            return p.name_with_prefix

    elif isinstance(p, ParamSeparator):
        return p.value

    else:
        raise TypeError(p)



# ruff: noqa: N802 N803
import typing as ta

from ... import check
from ...antlr import runtime as antlr4
from ...antlr.delimit import DelimitingLexer
from ...antlr.errors import SilentRaisingErrorListener
from .. import queries as no
from ._antlr.MinisqlLexer import MinisqlLexer  # type: ignore
from ._antlr.MinisqlParser import MinisqlParser  # type: ignore
from ._antlr.MinisqlVisitor import MinisqlVisitor  # type: ignore


##


class _ParseVisitor(MinisqlVisitor):
    def visit(self, ctx: antlr4.ParserRuleContext):
        check.isinstance(ctx, antlr4.ParserRuleContext)
        node = ctx.accept(self)
        return node

    def aggregateResult(self, aggregate, nextResult):
        if aggregate is not None:
            check.none(nextResult)
            return aggregate
        else:
            check.none(aggregate)
            return nextResult

    #

    def visitExprSelectItem(self, ctx: MinisqlParser.ExprSelectItemContext):
        value = self.visit(ctx.expr())
        label = self.visit(ctx.ident()) if ctx.ident() is not None else None
        return no.ExprSelectItem(value, label)

    def visitIntegerNumber(self, ctx: MinisqlParser.IntegerNumberContext):
        return no.Literal(int(ctx.INTEGER_VALUE().getText()))

    def visitPrimarySelect(self, ctx: MinisqlParser.PrimarySelectContext):
        items = [self.visit(i) for i in ctx.selectItem()]
        relations = [self.visit(r) for r in ctx.relation()]
        where = self.visit(ctx.where) if ctx.where is not None else None
        return no.Select(
            items=items,
            from_=check.single(relations) if relations else None,
            where=where,
        )

    def visitQuotedIdent(self, ctx: MinisqlParser.QuotedIdentContext):
        name = unquote(ctx.QUOTED_IDENT().getText(), '"')
        return no.Ident(name)

    def visitUnquotedIdent(self, ctx: MinisqlParser.UnquotedIdentContext):
        return no.Ident(ctx.getText())


##


def create_parser(buf: str) -> MinisqlParser:
    lexer = MinisqlLexer(antlr4.InputStream(buf))
    lexer.removeErrorListeners()
    lexer.addErrorListener(SilentRaisingErrorListener())

    stream = antlr4.CommonTokenStream(lexer)
    stream.fill()

    parser = MinisqlParser(stream)
    parser.removeErrorListeners()
    parser.addErrorListener(SilentRaisingErrorListener())

    return parser


##


def parse_stmt(buf: str, **kwargs) -> no.Stmt:
    parser = create_parser(buf, **kwargs)
    node = _ParseVisitor().visit(parser.singleStmt())
    return check.isinstance(node, no.Stmt)


class _DelimitingLexer(DelimitingLexer, MinisqlLexer):
    pass


def split_stmts(buf: str) -> ta.Sequence[str]:
    lexer = _DelimitingLexer(
        antlr4.InputStream(buf),
        delimiter_token=MinisqlParser.DELIMITER,
        delimiters=[';'],
    )
    lexer.removeErrorListeners()
    lexer.addErrorListener(SilentRaisingErrorListener())

    lst, part = lexer.split()
    if part.strip():
        raise ValueError(part)

    return [s for s, _ in lst]


def parse_stmts(buf: str, **kwargs) -> ta.Sequence[no.Stmt]:
    return [parse_stmt(sb, **kwargs) for sb in split_stmts(buf)]


##


def quote(val: str, q: str) -> str:
    return q + val.replace(q, q * 2) + q


def unquote(val: str, q: str) -> str:
    check.arg(val.startswith(q) and val.endswith(q))
    return val[1:-1].replace(q * 2, q)



"""
TODO:
 - cache kwarg_keys
 - tag annotations? x: ta.Annotated[int, inj.Tag('foo')]
 - tag decorator - @inj.tag(x='foo')
 - *unpack optional here*
"""
import dataclasses as dc
import inspect
import typing as ta
import weakref

from ... import check
from ... import reflect as rfl
from ..errors import ConflictingKeyError
from ..inspect import Kwarg
from ..inspect import KwargsTarget
from ..keys import Key
from ..types import Tag


T = ta.TypeVar('T')
P = ta.ParamSpec('P')
R = ta.TypeVar('R')


##


_SIGNATURE_CACHE: ta.MutableMapping[ta.Any, inspect.Signature] = weakref.WeakKeyDictionary()


def signature(obj: ta.Any) -> inspect.Signature:
    def get() -> inspect.Signature:
        return inspect.signature(obj, eval_str=True)
    try:
        return _SIGNATURE_CACHE[obj]
    except TypeError:
        return get()
    except KeyError:
        pass
    sig = get()
    _SIGNATURE_CACHE[obj] = sig
    return sig


##


_TAGS: ta.MutableMapping[ta.Any, dict[str, ta.Any]] = weakref.WeakKeyDictionary()


def tag(obj: T, **kwargs: ta.Any) -> T:
    for v in kwargs.values():
        if isinstance(v, Tag):
            raise TypeError(v)
    _TAGS.setdefault(obj, {}).update(**kwargs)
    return obj


def tags(**kwargs: ta.Any) -> ta.Callable[[ta.Callable[P, R]], ta.Callable[P, R]]:
    def inner(obj):
        _TAGS[obj] = kwargs
        return obj
    return inner


##


def build_kwargs_target(
        obj: ta.Any,
        *,
        skip_args: int = 0,
        skip_kwargs: ta.Iterable[str] | None = None,
        raw_optional: bool = False,
) -> KwargsTarget:
    sig = signature(obj)
    tags = _TAGS.get(obj)

    skip_names: set[str] = set()
    if skip_kwargs is not None:
        skip_names.update(check.not_isinstance(skip_kwargs, str))

    seen: set[Key] = set()
    kws: list[Kwarg] = []
    for p in list(sig.parameters.values())[skip_args:]:
        if p.name in skip_names:
            continue

        if p.annotation is inspect.Signature.empty:
            if p.default is not inspect.Parameter.empty:
                raise KeyError(f'{obj}, {p.name}')
            continue

        if p.kind not in (inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY):
            raise TypeError(sig)

        ann = p.annotation
        if (
                not raw_optional and
                isinstance(rf := rfl.type_(ann), rfl.Union) and
                rf.is_optional
        ):
            ann = rf.without_none()

        rty = rfl.type_(ann)

        tag = None
        if isinstance(rty, rfl.Annotated):
            for e in rty.md:
                if isinstance(e, Tag):
                    tag = e.tag
                    break

        k: Key = Key(rfl.strip_annotations(rty), tag=tag)
        if tags is not None and (pt := tags.get(p.name)) is not None:
            k = dc.replace(k, tag=pt)

        if k in seen:
            raise ConflictingKeyError(k)
        seen.add(k)

        kws.append(Kwarg(
            p.name,
            k,
            p.default is not inspect.Parameter.empty,
        ))

    return KwargsTarget(
        obj,
        kws,
    )



import typing as ta

from .lex import SCALAR_VALUE_TYPES
from .parse import BeginArray
from .parse import BeginObject
from .parse import EndArray
from .parse import EndObject
from .parse import JsonStreamObject
from .parse import JsonStreamParserEvent
from .parse import Key


##


class JsonObjectBuilder:
    def __init__(
            self,
            *,
            yield_object_lists: bool = False,
    ) -> None:
        super().__init__()

        self._stack: list[JsonStreamObject | list | Key] = []
        self._yield_object_lists = yield_object_lists

    class StateError(Exception):
        pass

    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.close()

    def close(self) -> None:
        if self._stack:
            raise self.StateError

    def _emit_value(self, v: ta.Any) -> ta.Iterable[ta.Any]:
        if not (stk := self._stack):
            return (v,)

        tv = stk[-1]
        if isinstance(tv, Key):
            stk.pop()
            if not stk:
                raise self.StateError

            tv2 = stk[-1]
            if not isinstance(tv2, JsonStreamObject):
                raise self.StateError

            tv2.append((tv.key, v))
            return ()

        elif isinstance(tv, list):
            tv.append(v)
            return ()

        else:
            raise self.StateError

    def __call__(self, e: JsonStreamParserEvent) -> ta.Iterable[ta.Any]:
        stk = self._stack

        #

        if isinstance(e, SCALAR_VALUE_TYPES):
            return self._emit_value(e)

        #

        elif e is BeginObject:
            stk.append(JsonStreamObject())
            return ()

        elif isinstance(e, Key):
            if not stk or not isinstance(stk[-1], JsonStreamObject):
                raise self.StateError

            stk.append(e)
            return ()

        elif e is EndObject:
            tv: ta.Any
            if not stk or not isinstance(tv := stk.pop(), JsonStreamObject):
                raise self.StateError

            if not self._yield_object_lists:
                tv = dict(tv)

            return self._emit_value(tv)

        #

        elif e is BeginArray:
            stk.append([])
            return ()

        elif e is EndArray:
            if not stk or not isinstance(tv := stk.pop(), list):
                raise self.StateError

            return self._emit_value(tv)

        #

        else:
            raise TypeError(e)



# ruff: noqa: I001
from .flattening import (  # noqa
    ConfigFlattening as Flattening,
)

from .inheritance import (  # noqa
    build_config_inherited_values as build_inherited_values,
)

from .matching import (  # noqa
    MatchingConfigRewriter as MatchingRewriter,

    matched_config_rewrite as matched_rewrite,
)

from .merging import (  # noqa
    merge_configs as merge,
)

from .names import (  # noqa
    build_config_named_children as build_named_children,
)

from .rewriting import (  # noqa
    ConfigRewriterItem as RewriterItem,
    ConfigRewriterPath as RewriterPath,

    RawConfigMetadata as RawMetadata,

    ConfigRewriter as Rewriter,
)

from .strings import (  # noqa
    StringConfigRewriter as StringRewriter,

    format_config_strings as format_strings,
)



##
# stdlib interface

from dataclasses import (  # noqa
    FrozenInstanceError,

    MISSING,
    KW_ONLY,

    InitVar,
    Field,

    field,

    dataclass,
    make_dataclass,

    fields,

    is_dataclass,
)

from .api import (  # noqa
    dataclass as xdataclass,

    make_dataclass as xmake_dataclass,

    field as xfield,
)

from .tools.as_ import (  # noqa
    asdict,
    astuple,
)

from .concerns.replace import (  # noqa
    replace,
)


##
# globals hack

globals()['field'] = xfield

globals()['dataclass'] = xdataclass
globals()['make_dataclass'] = xmake_dataclass


##
# additional interface

from .api import (  # noqa
    append_class_metadata,
    extra_class_params,
    init,
    metadata,
    validate,

    extra_field_params,
    set_field_metadata,
    update_extra_field_params,
)

from .errors import (  # noqa
    FieldFnValidationError,
    FieldTypeValidationError,
    FieldValidationError,
    FnValidationError,
    TypeValidationError,
    ValidationError,
)

from .metaclass.bases import (  # noqa
    Box,
    Case,
    Data,
    Frozen,
)

from .metaclass.meta import (  # noqa
    DataMeta,
)

from .metaclass.specs import (  # noqa
    get_metaclass_spec,
)

from .reflection import (  # noqa
    reflect,
)

from .tools.as_ import (  # noqa
    shallow_asdict,
    shallow_astuple,
)

from .tools.iter import (  # noqa
    fields_dict,

    iter_items,
    iter_keys,
    iter_values,
)

from .tools.modifiers import (  # noqa
    field_modifier,
    update_fields,
)

from .tools.replace import (  # noqa
    deep_replace,
)

from .tools.repr import (  # noqa
    opt_repr,
    truthy_repr,
)

from .tools.static import (  # noqa
    Static,
)


##
# lite imports

from ..lite.dataclasses import (  # noqa
    is_immediate_dataclass,

    dataclass_maybe_post_init as maybe_post_init,
)



import dataclasses as dc
import typing as ta

from .... import check
from .... import lang
from ...debug import DEBUG
from ...specs import FieldSpec
from ...utils import chain_mapping_proxy


##


class _ExtraFieldParamsMetadata(lang.Marker):
    pass


def extra_field_params(**kwargs: ta.Any) -> ta.Mapping[ta.Any, ta.Any]:
    return {_ExtraFieldParamsMetadata: kwargs}


##


def get_field_spec(f: dc.Field) -> FieldSpec | None:
    try:
        fs = f.metadata[FieldSpec]
    except KeyError:
        return None

    return check.isinstance(fs, FieldSpec)


##


def set_field_metadata(
        f: dc.Field,
        metadata: ta.Mapping[ta.Any, ta.Any],
) -> dc.Field:
    if DEBUG:
        check.isinstance(f, dc.Field)

    md: ta.Any = f.metadata

    mdu: dict = {}
    for k, v in metadata.items():
        if md is None or md.get(k) != v:
            mdu[k] = v  # noqa
    if not mdu:
        return f

    if md is None:
        ms = [mdu]
    else:
        ms = [mdu, md]

    f.metadata = chain_mapping_proxy(*ms)
    return f


#


def set_field_spec_metadata(
        f: dc.Field,
        fs: FieldSpec,
) -> None:
    set_field_metadata(
        f,
        {
            FieldSpec: fs,
            _ExtraFieldParamsMetadata: {},
        },
    )


##


def update_extra_field_params(
        f: dc.Field,
        *,
        unless_non_default: bool = False,
        **kwargs: ta.Any,
) -> dc.Field:
    fe = f.metadata.get(_ExtraFieldParamsMetadata, {})
    return set_field_metadata(f, {
        _ExtraFieldParamsMetadata: {
            **(fe if not unless_non_default else {}),
            **kwargs,
            **(fe if unless_non_default else {}),
        },
    })



from ... import lang
from ... import marshal as msh
from .dtypes import Dtype
from .elements import Element


def _install_poly(cls: type) -> None:
    p = msh.polymorphism_from_subclasses(cls, naming=msh.Naming.SNAKE)
    msh.install_standard_factories(
        msh.PolymorphismMarshalerFactory(p),
        msh.PolymorphismUnmarshalerFactory(p),
    )


@lang.static_init
def _install_standard_marshalling() -> None:
    _install_poly(Dtype)
    _install_poly(Element)



import types
import typing as ta

from ... import dataclasses as dc
from ... import lang


##


Value: ta.TypeAlias = ta.Any


##


class NodeComparisonTypeError(TypeError):
    def __init__(self, cls: type['Node'], *args: ta.Any) -> None:
        super().__init__(
            'Query node types are not comparable or hashable - this would be a common source of confusion when '
            'constructing query filters',
            cls,
            *args,
        )


class Node(
    dc.Frozen,
    lang.Abstract,
    eq=False,
    confer=frozenset([
        *dc.get_metaclass_spec(dc.Frozen).confer,
        'eq',
    ]),
):
    @ta.final
    def __hash__(self) -> ta.NoReturn:
        raise NodeComparisonTypeError(type(self))

    @ta.final
    def __eq__(self, other) -> ta.NoReturn:
        raise NodeComparisonTypeError(type(self))

    @ta.final
    def __ne__(self, other) -> ta.NoReturn:
        raise NodeComparisonTypeError(type(self))

    #

    @dc.dataclass(frozen=True)
    class _Fields:
        cmp_fields: ta.Sequence[str]
        hash_fields: ta.Sequence[str]

    __node_fields__: ta.ClassVar[tuple[str, ...]]

    @classmethod
    def _fields(cls) -> _Fields:
        try:
            return cls.__dict__['__node_fields__']
        except KeyError:
            pass

        dc_rfl = dc.reflect(cls)
        fields = Node._Fields(
            cmp_fields=tuple(f.name for f in dc_rfl.instance_fields if f.compare),
            hash_fields=tuple(f.name for f in dc_rfl.instance_fields if (f.compare if f.hash is None else f.hash)),
        )

        setattr(cls, '__node_fields__', fields)
        return fields

    _hash: ta.ClassVar[int]

    def hash(self) -> int:
        try:
            return self._hash
        except AttributeError:
            pass

        h = hash(tuple(getattr(self, f) for f in self._fields().hash_fields))
        object.__setattr__(self, '_hash', h)
        return h

    def eq(self, other: 'Node') -> bool | types.NotImplementedType:
        if self is other:
            return True

        def rec(l, r):
            if l.__class__ is not r.__class__:
                return NotImplemented

            if isinstance(l, lang.BUILTIN_SCALAR_ITERABLE_TYPES):
                return l == r

            elif isinstance(l, ta.Mapping):
                ks = set(l)
                if set(r) != ks:
                    return False
                for k, lv in l.items():
                    rv = r[k]
                    if (ret := rec(lv, rv)) is not True:
                        return ret
                return True

            elif isinstance(l, ta.Sequence):
                if len(l) != len(r):
                    return False
                for le, re in zip(l, r):
                    if (ret := rec(le, re)) is not True:
                        return ret
                return True

            elif isinstance(l, Node):
                return l.eq(r)

            else:
                return l == r

        for f in self._fields().cmp_fields:
            lf = getattr(self, f)
            rf = getattr(other, f)
            if (ret := rec(lf, rf)) is not True:
                return ret
        return True


##


class Builder(lang.Abstract):
    pass



import typing as ta

from .lite.check import CheckArgsRenderer as ArgsRenderer  # noqa
from .lite.check import CheckExceptionFactory as ExceptionFactory  # noqa
from .lite.check import CheckLateConfigureFn as LateConfigureFn  # noqa
from .lite.check import CheckMessage as Message  # noqa
from .lite.check import CheckOnRaiseFn as OnRaiseFn  # noqa
from .lite.check import Checks
from .lite.check import check


_isinstance = isinstance
_issubclass = issubclass
_callable = callable


##


register_on_raise = check.register_on_raise
unregister_on_raise = check.unregister_on_raise


##


def _try_get_args_rendering() -> ArgsRenderer | None:
    try:
        from .diag.asts import ArgsRenderer

        ArgsRenderer.smoketest()

    except Exception:  # noqa
        return None

    def _real_render_args(fmt: str, *args: ta.Any) -> str | None:
        ra = ArgsRenderer(back=3).render_args(*args)
        if ra is None:
            return None

        return fmt % tuple(str(a) for a in ra)

    return _real_render_args


def _try_enable_args_rendering(c: Checks) -> None:
    if (rf := _try_get_args_rendering()) is not None:
        c.set_args_renderer(rf)


check.register_late_configure(_try_enable_args_rendering)


##


isinstance = check.isinstance  # noqa
of_isinstance = check.of_isinstance
cast = check.cast
of_cast = check.of_cast
not_isinstance = check.not_isinstance
of_not_isinstance = check.of_not_isinstance

#

issubclass = check.issubclass  # noqa
not_issubclass = check.not_issubclass

#

in_ = check.in_
not_in = check.not_in
empty = check.empty
iterempty = check.iterempty
not_empty = check.not_empty
unique = check.unique
single = check.single
opt_single = check.opt_single

#

none = check.none
not_none = check.not_none

#

equal = check.equal
not_equal = check.not_equal
is_ = check.is_
is_not = check.is_not
callable = check.callable  # noqa
non_empty_str = check.non_empty_str
replacing = check.replacing
replacing_none = check.replacing_none

#

arg = check.arg
state = check.state



import abc
import typing as ta

from .. import lang
from .accessor import TypedValuesAccessor
from .collection import TypedValues
from .generic import TypedValueGeneric
from .values import TypedValue


TypedValueT = ta.TypeVar('TypedValueT', bound=TypedValue)


##


class TypedValueHolder(
    TypedValuesAccessor[TypedValueT],
    TypedValueGeneric[TypedValueT],
    lang.Abstract,
):
    @property
    @abc.abstractmethod
    def _typed_values(self) -> TypedValues[TypedValueT] | None:
        raise NotImplementedError

    #

    def _typed_value_contains(self, cls):
        if (tvs := self._typed_values) is not None:
            return cls in tvs
        return False

    def _typed_value_getitem(self, key):
        if (tvs := self._typed_values) is not None:
            return tvs[key]
        if isinstance(key, int):
            raise IndexError(key)
        elif isinstance(key, type):
            raise KeyError(key)
        else:
            raise TypeError(key)

    def _typed_value_get_any(self, cls):
        if (tvs := self._typed_values) is not None:
            return tvs.get_any(cls)
        return []



import abc
import typing as ta


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class OverweightError(Exception):
    pass


Eviction: ta.TypeAlias = ta.Callable[['Cache'], None]


class Cache(ta.MutableMapping[K, V]):
    """
    https://google.github.io/guava/releases/16.0/api/docs/com/google/common/cache/CacheBuilder.html
    """

    @abc.abstractmethod
    def reap(self) -> None:
        pass

    class Stats(ta.NamedTuple):
        seq: int
        size: int
        weight: float
        hits: int
        misses: int
        max_size_ever: int
        max_weight_ever: float

    @property
    @abc.abstractmethod
    def stats(self) -> Stats:
        raise NotImplementedError



import typing as ta

from .... import dataclasses as dc
from .... import lang
from .base import Keyword


##


@dc.dataclass(frozen=True)
class UnknownKeyword(Keyword, lang.Final):
    tag: str  # type: ignore[misc]
    value: ta.Any



import functools
import typing as ta

import anyio.abc
import sniffio

from ... import lang


P = ta.ParamSpec('P')
T = ta.TypeVar('T')


##


async def eof_to_empty(fn: ta.Callable[..., ta.Awaitable[T]], *args: ta.Any, **kwargs: ta.Any) -> T | bytes:
    try:
        return await fn(*args, **kwargs)
    except anyio.EndOfStream:
        return b''


async def gather(*fns: ta.Callable[..., ta.Awaitable[T]], take_first: bool = False) -> list[lang.Maybe[T]]:
    results: list[lang.Maybe[T]] = [lang.empty()] * len(fns)

    async def inner(fn, i):
        results[i] = lang.just(await fn())
        if take_first:
            tg.cancel_scope.cancel()

    async with anyio.create_task_group() as tg:
        for i, fn in enumerate(fns):
            tg.start_soon(inner, fn, i)

    return results


async def first(*fns: ta.Callable[..., ta.Awaitable[T]], **kwargs: ta.Any) -> list[lang.Maybe[T]]:
    return await gather(*fns, take_first=True, **kwargs)


##


def get_current_task() -> anyio.TaskInfo | None:
    try:
        return anyio.get_current_task()
    except sniffio.AsyncLibraryNotFoundError:
        return None


##


async def call_with_task_group(
        fn: ta.Callable[ta.Concatenate[anyio.abc.TaskGroup, P], ta.Awaitable[T]],
        *args: ta.Any,
        **kwargs: ta.Any,
) -> T:
    async with anyio.create_task_group() as tg:
        return await fn(tg, *args, **kwargs)


def run_with_task_group(
        fn: ta.Callable[ta.Concatenate[anyio.abc.TaskGroup, P], ta.Awaitable[T]],
        *args: ta.Any,
        **kwargs: ta.Any,
) -> T:
    return anyio.run(
        functools.partial(call_with_task_group, fn, *args),
        **kwargs,
    )



# type: ignore
# ruff: noqa
# flake8: noqa
# Generated from Minisql.g4 by ANTLR 4.13.2
from omlish.antlr._runtime._all import *
if "." in __name__:
    from .MinisqlParser import MinisqlParser
else:
    from MinisqlParser import MinisqlParser

# This class defines a complete generic visitor for a parse tree produced by MinisqlParser.

class MinisqlVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by MinisqlParser#singleStmt.
    def visitSingleStmt(self, ctx:MinisqlParser.SingleStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#select.
    def visitSelect(self, ctx:MinisqlParser.SelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#cteSelect.
    def visitCteSelect(self, ctx:MinisqlParser.CteSelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#cte.
    def visitCte(self, ctx:MinisqlParser.CteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#unionSelect.
    def visitUnionSelect(self, ctx:MinisqlParser.UnionSelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#unionItem.
    def visitUnionItem(self, ctx:MinisqlParser.UnionItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#primarySelect.
    def visitPrimarySelect(self, ctx:MinisqlParser.PrimarySelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#allSelectItem.
    def visitAllSelectItem(self, ctx:MinisqlParser.AllSelectItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#exprSelectItem.
    def visitExprSelectItem(self, ctx:MinisqlParser.ExprSelectItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#expr.
    def visitExpr(self, ctx:MinisqlParser.ExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#unaryBooleanExpr.
    def visitUnaryBooleanExpr(self, ctx:MinisqlParser.UnaryBooleanExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#predicatedBooleanExpr.
    def visitPredicatedBooleanExpr(self, ctx:MinisqlParser.PredicatedBooleanExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#binaryBooleanExpr.
    def visitBinaryBooleanExpr(self, ctx:MinisqlParser.BinaryBooleanExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#castBooleanExpr.
    def visitCastBooleanExpr(self, ctx:MinisqlParser.CastBooleanExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#cmpPredicate.
    def visitCmpPredicate(self, ctx:MinisqlParser.CmpPredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#isNullPredicate.
    def visitIsNullPredicate(self, ctx:MinisqlParser.IsNullPredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#inListPredicate.
    def visitInListPredicate(self, ctx:MinisqlParser.InListPredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#inSelectPredicate.
    def visitInSelectPredicate(self, ctx:MinisqlParser.InSelectPredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#likePredicate.
    def visitLikePredicate(self, ctx:MinisqlParser.LikePredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#arithValueExpr.
    def visitArithValueExpr(self, ctx:MinisqlParser.ArithValueExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#unaryValueExpr.
    def visitUnaryValueExpr(self, ctx:MinisqlParser.UnaryValueExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#primaryValueExpr.
    def visitPrimaryValueExpr(self, ctx:MinisqlParser.PrimaryValueExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#functionCallExpr.
    def visitFunctionCallExpr(self, ctx:MinisqlParser.FunctionCallExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#starFunctionCallExpr.
    def visitStarFunctionCallExpr(self, ctx:MinisqlParser.StarFunctionCallExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#caseExpr.
    def visitCaseExpr(self, ctx:MinisqlParser.CaseExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#selectExpr.
    def visitSelectExpr(self, ctx:MinisqlParser.SelectExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#parenExpr.
    def visitParenExpr(self, ctx:MinisqlParser.ParenExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#simplePrimaryExpr.
    def visitSimplePrimaryExpr(self, ctx:MinisqlParser.SimplePrimaryExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#simpleExpr.
    def visitSimpleExpr(self, ctx:MinisqlParser.SimpleExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#caseItem.
    def visitCaseItem(self, ctx:MinisqlParser.CaseItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#over.
    def visitOver(self, ctx:MinisqlParser.OverContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#sortItem.
    def visitSortItem(self, ctx:MinisqlParser.SortItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#aliasedRelation.
    def visitAliasedRelation(self, ctx:MinisqlParser.AliasedRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#joinRelation.
    def visitJoinRelation(self, ctx:MinisqlParser.JoinRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#selectRelation.
    def visitSelectRelation(self, ctx:MinisqlParser.SelectRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#tableRelation.
    def visitTableRelation(self, ctx:MinisqlParser.TableRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#parenRelation.
    def visitParenRelation(self, ctx:MinisqlParser.ParenRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#groupBy.
    def visitGroupBy(self, ctx:MinisqlParser.GroupByContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#qualifiedName.
    def visitQualifiedName(self, ctx:MinisqlParser.QualifiedNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#identList.
    def visitIdentList(self, ctx:MinisqlParser.IdentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#ident.
    def visitIdent(self, ctx:MinisqlParser.IdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#quotedIdent.
    def visitQuotedIdent(self, ctx:MinisqlParser.QuotedIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#integerNumber.
    def visitIntegerNumber(self, ctx:MinisqlParser.IntegerNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#decimalNumber.
    def visitDecimalNumber(self, ctx:MinisqlParser.DecimalNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#floatNumber.
    def visitFloatNumber(self, ctx:MinisqlParser.FloatNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#string.
    def visitString(self, ctx:MinisqlParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#null.
    def visitNull(self, ctx:MinisqlParser.NullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#true.
    def visitTrue(self, ctx:MinisqlParser.TrueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#false.
    def visitFalse(self, ctx:MinisqlParser.FalseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#setQuantifier.
    def visitSetQuantifier(self, ctx:MinisqlParser.SetQuantifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#joinType.
    def visitJoinType(self, ctx:MinisqlParser.JoinTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#cmpOp.
    def visitCmpOp(self, ctx:MinisqlParser.CmpOpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#arithOp.
    def visitArithOp(self, ctx:MinisqlParser.ArithOpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#unaryOp.
    def visitUnaryOp(self, ctx:MinisqlParser.UnaryOpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by MinisqlParser#unquotedIdent.
    def visitUnquotedIdent(self, ctx:MinisqlParser.UnquotedIdentContext):
        return self.visitChildren(ctx)



del MinisqlParser


import dataclasses as dc
import typing as ta

from ..internals import STD_FIELDS_ATTR
from ..internals import StdFieldType
from ..internals import std_field_type
from ..internals import std_is_dataclass_instance
from ..processing.base import Processor
from ..processing.priority import ProcessorPriority
from ..processing.registry import register_processor_type
from ..utils import set_new_attribute


T = ta.TypeVar('T')


##


def replace(obj, /, **changes):  # noqa
    if not std_is_dataclass_instance(obj):
        raise TypeError('replace() should be called on dataclass instances')
    return _replace(obj, **changes)


def _replace(obj, /, **changes):
    for f in getattr(obj, STD_FIELDS_ATTR).values():
        if (ft := std_field_type(f)) is StdFieldType.CLASS_VAR:
            continue

        if not f.init:
            if f.name in changes:
                raise TypeError(f'field {f.name} is declared with init=False, it cannot be specified with replace()')
            continue

        if f.name not in changes:
            if ft is StdFieldType.INIT_VAR and f.default is dc.MISSING:
                raise TypeError(f'InitVar {f.name!r} must be specified with replace()')
            changes[f.name] = getattr(obj, f.name)

    return obj.__class__(**changes)


@register_processor_type(priority=ProcessorPriority.POST_GENERATION)
class ReplaceProcessor(Processor):
    def process(self, cls: type) -> type:
        set_new_attribute(cls, '__replace__', _replace)
        return cls



"""
https://www.jsonrpc.org/specification

TODO:
 - drop NotSpecified, use lang.Maybe, make marshal do that
 - server/client impl

See:
 - https://github.com/python-lsp/python-lsp-jsonrpc
"""
import operator
import types
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh


NUMBER_TYPES: tuple[type, ...] = (int, float)
Number: ta.TypeAlias = int | float

Object: ta.TypeAlias = ta.Mapping[str, ta.Any]

ID_TYPES: tuple[type, ...] = (str, *NUMBER_TYPES, types.NoneType)
Id: ta.TypeAlias = str | Number | None


##


VERSION = '2.0'


##


class NotSpecified(lang.Marker):
    pass


def is_not_specified(v: ta.Any) -> bool:
    return v is NotSpecified


##


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['id'], omit_if=is_not_specified, default=lang.just(NotSpecified))
@msh.update_fields_metadata(['params'], omit_if=operator.not_)
class Request(lang.Final):
    id: Id | type[NotSpecified]

    @property
    def is_notification(self) -> bool:
        return self.id is NotSpecified

    def id_value(self) -> Id:
        return check.isinstance(self.id, ID_TYPES)

    method: str
    params: Object | None = None

    jsonrpc: str = dc.field(default=VERSION, kw_only=True)
    dc.validate(lambda self: self.jsonrpc == VERSION)



def request(id: Id, method: str, params: Object | None = None) -> Request:  # noqa
    return Request(id, method, params)


def notification(method: str, params: Object | None = None) -> Request:
    return Request(NotSpecified, method, params)


##


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['result', 'error'], omit_if=is_not_specified)
class Response(lang.Final):
    id: Id
    dc.validate(lambda self: self.id is not NotSpecified)

    _: dc.KW_ONLY

    result: ta.Any = dc.field(default=NotSpecified)
    error: ta.Union['Error', type[NotSpecified]] = dc.field(default=NotSpecified)
    dc.validate(lambda self: is_not_specified(self.result) ^ is_not_specified(self.error))

    jsonrpc: str = dc.field(default=VERSION)
    dc.validate(lambda self: self.jsonrpc == VERSION)


def result(id: Id, result: ta.Any) -> Response:  # noqa
    return Response(id, result=result)


##


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['data'], omit_if=is_not_specified)
class Error(lang.Final):
    code: int
    message: str
    data: ta.Any = NotSpecified


def error(id: Id, error: Error) -> Response:  # noqa
    return Response(id, error=error)


##


Message: ta.TypeAlias = Request | Response


def detect_message_type(dct: ta.Mapping[str, ta.Any]) -> type[Message]:
    if 'method' in dct:
        return Request
    else:
        return Response



from .keywords.base import (  # noqa
    Keyword,
    Keywords,
    KnownKeyword,
)

from .keywords.core import (  # noqa
    CoreKeyword,
    Defs,
    Id,
    Ref,
    SchemaKeyword,
)

from .keywords.format import (  # noqa
    Format,
    FormatKeyword,
)

from .keywords.metadata import (  # noqa
    Description,
    MetadataKeyword,
    Title,
)

from .keywords.parse import (  # noqa
    DEFAULT_KEYWORD_SUPERTYPES,
    DEFAULT_KEYWORD_TYPES,
    DEFAULT_KEYWORD_TYPES_BY_TAG,
    DEFAULT_KEYWORD_PARSER,
    KeywordParser,
    build_keyword_types_by_tag,
    parse_keyword,
    parse_keywords,
)

from .keywords.render import (  # noqa
    render_keyword,
    render_keywords,
)

from .keywords.unknown import (  # noqa
    UnknownKeyword,
)

from .keywords.validation import (  # noqa
    ExclusiveMaximum,
    ExclusiveMinimum,
    Items,
    MaxItems,
    Maximum,
    MinItems,
    Minimum,
    Properties,
    Required,
    Type,
    UniqueItems,
    ValidationKeyword,
)

from .types import (  # noqa
    JsonType,
    TYPE_SETS_BY_JSON_TYPE,
)



"""
from_json(
    json_data: bytes,
    /,
    *,
    allow_inf_nan: bool = True,
    cache_mode: Literal[True, False, "all", "keys", "none"] = "all",
    partial_mode: Literal[True, False, "off", "on", "trailing-strings"] = False,
    catch_duplicate_keys: bool = False,
    lossless_floats: bool = False,
) -> Any:
    json_data: The JSON data to parse
    allow_inf_nan: Whether to allow infinity (`Infinity` an `-Infinity`) and `NaN` values to float fields.
    Defaults to True.
    cache_mode: cache Python strings to improve performance at the cost of some memory usage
    - True / 'all' - cache all strings
    - 'keys' - cache only object keys
    - False / 'none' - cache nothing
    partial_mode: How to handle incomplete strings:
    - False / 'off' - raise an exception if the input is incomplete
    - True / 'on' - allow incomplete JSON but discard the last string if it is incomplete
    - 'trailing-strings' - allow incomplete JSON, and include the last incomplete string in the output
    catch_duplicate_keys: if True, raise an exception if objects contain the same key multiple times
    lossless_floats: if True, preserve full detail on floats using `LosslessFloat`
"""



import sqlalchemy as sa
import sqlalchemy.ext.compiler


class paren(sa.sql.expression.UnaryExpression):  # noqa
    __visit_name__ = 'paren'
    inherit_cache = True


@sa.ext.compiler.compiles(paren)
def _compile_paren(element, compiler, **kw):
    return '(%s)' % (element.element._compiler_dispatch(compiler),)  # noqa



from .secrets import (  # noqa
    CachingSecrets,
    CompositeSecrets,
    EMPTY_SECRETS,
    EmptySecrets,
    EnvVarSecrets,
    FnSecrets,
    LoggingSecrets,
    MappingSecrets,
    Secret,
    SecretRef,
    SecretRefOrStr,
    Secrets,
    secret_field,
    secret_repr,
)

ref = SecretRef


##


from ..lang.imports import _register_conditional_import  # noqa

_register_conditional_import('..marshal', '.marshal', __package__)



# ruff: noqa: UP007
# @omlish-lite
import contextlib
import os
import threading
import typing as ta

from ...lite.check import check
from ..forkhooks import ForkHook
from .pidfile import Pidfile


##


class _PidfileManager(ForkHook):
    """
    Manager for controlled inheritance of Pidfiles across forks in the presence of multiple threads. There is of course
    no safe or correct way to mix the use of fork and multiple active threads, and one should never write code which
    does so, but in the Real World one may still find oneself in such a situation outside of their control (such as when
    running under Pycharm's debugger which forces the use of forked multiprocessing).
    """

    _lock: ta.ClassVar[threading.Lock] = threading.Lock()
    _pidfile_threads: ta.ClassVar[ta.MutableMapping[Pidfile, threading.Thread]] = {}
    _num_kills: ta.ClassVar[int] = 0

    @classmethod
    def num_kills(cls) -> int:
        return cls._num_kills

    @classmethod
    def _after_fork_in_child(cls) -> None:
        with cls._lock:
            th = threading.current_thread()
            for pf, pf_th in list(cls._pidfile_threads.items()):
                if pf_th is not th:
                    pf.close()
                    del cls._pidfile_threads[pf]
                    cls._num_kills += 1

    @classmethod
    @contextlib.contextmanager
    def inheritable_pidfile_context(
            cls,
            path: str,
            *,
            inheritable: bool = True,
            **kwargs: ta.Any,
    ) -> ta.Iterator[Pidfile]:
        """
        A contextmanager for creating and managing a Pidfile which will only be inherited by forks of the calling /
        creating thread.
        """

        check.arg(inheritable)

        cls.install()

        pf = Pidfile(
            path,
            inheritable=False,
            **kwargs,
        )

        with cls._lock:
            cls._pidfile_threads[pf] = threading.current_thread()

        try:
            with pf:
                os.set_inheritable(check.not_none(pf.fileno()), True)
                yield pf

        finally:
            with cls._lock:
                del cls._pidfile_threads[pf]


def open_inheritable_pidfile(path: str, **kwargs: ta.Any) -> ta.ContextManager[Pidfile]:
    return _PidfileManager.inheritable_pidfile_context(path, **kwargs)  # noqa



import typing as ta

from .... import check
from .... import collections as col
from .... import lang
from .base import BooleanKeyword
from .base import Keyword
from .base import Keywords
from .base import KeywordsKeyword
from .base import KnownKeyword
from .base import NumberKeyword
from .base import StrKeyword
from .base import StrOrStrsKeyword
from .base import StrToKeywordsKeyword
from .core import CoreKeyword
from .format import FormatKeyword
from .metadata import MetadataKeyword
from .unknown import UnknownKeyword
from .validation import ValidationKeyword


KeywordT = ta.TypeVar('KeywordT', bound=Keyword)


##


def build_keyword_types_by_tag(keyword_types: ta.Iterable[type[KnownKeyword]]) -> ta.Mapping[str, type[KnownKeyword]]:
    return col.make_map(((t, kt) for kt in keyword_types for t in kt.tag_and_aliases), strict=True)


DEFAULT_KEYWORD_SUPERTYPES: ta.AbstractSet[type[KnownKeyword]] = frozenset([
    CoreKeyword,
    FormatKeyword,
    MetadataKeyword,
    ValidationKeyword,
])

DEFAULT_KEYWORD_TYPES: ta.AbstractSet[type[KnownKeyword]] = frozenset(lang.flatten(
    lang.deep_subclasses(st, concrete_only=True) for st in DEFAULT_KEYWORD_SUPERTYPES
))

DEFAULT_KEYWORD_TYPES_BY_TAG: ta.Mapping[str, type[KnownKeyword]] = build_keyword_types_by_tag(DEFAULT_KEYWORD_TYPES)


##


class KeywordParser:
    def __init__(
            self,
            *,
            keyword_types: ta.Union[  # noqa
                ta.Iterable[type[KnownKeyword]],
                ta.Mapping[str, type[KnownKeyword]],
                None,
            ] = None,
            allow_unknown: bool = False,
    ) -> None:
        super().__init__()

        if keyword_types is None:
            self._keyword_types_by_tag = DEFAULT_KEYWORD_TYPES_BY_TAG
        elif isinstance(keyword_types, ta.Mapping):
            self._keyword_types_by_tag = keyword_types
        else:
            self._keyword_types_by_tag = build_keyword_types_by_tag(keyword_types)

        self._allow_unknown = allow_unknown

    def parse_keyword(self, cls: type[KeywordT], v: ta.Any) -> KeywordT:
        if issubclass(cls, BooleanKeyword):
            return cls(check.isinstance(v, bool))  # type: ignore

        elif issubclass(cls, NumberKeyword):
            return cls(check.isinstance(v, (int, float)))  # type: ignore

        elif issubclass(cls, StrKeyword):
            return cls(check.isinstance(v, str))  # type: ignore

        elif issubclass(cls, StrOrStrsKeyword):
            ss: str | ta.Sequence[str]
            if isinstance(v, str):
                ss = v
            elif isinstance(v, ta.Iterable):
                ss = col.seq_of(check.of_isinstance(str))(v)
            else:
                raise TypeError(v)
            return cls(ss)  # type: ignore

        elif issubclass(cls, KeywordsKeyword):
            return cls(self.parse_keywords(v))  # type: ignore

        elif issubclass(cls, StrToKeywordsKeyword):
            return cls({k: self.parse_keywords(mv) for k, mv in v.items()})  # type: ignore

        else:
            raise TypeError(cls)

    def parse_keywords(self, dct: ta.Mapping[str, ta.Any]) -> Keywords:
        lst: list[Keyword] = []

        for k, v in dct.items():
            try:
                cls = self._keyword_types_by_tag[k]

            except KeyError:
                if not self._allow_unknown:
                    raise

                lst.append(UnknownKeyword(k, v))

            else:
                lst.append(self.parse_keyword(cls, v))

        return Keywords(lst)


##


DEFAULT_KEYWORD_PARSER = KeywordParser()

parse_keyword = DEFAULT_KEYWORD_PARSER.parse_keyword
parse_keywords = DEFAULT_KEYWORD_PARSER.parse_keywords



from .base import (  # noqa
    Closer,
    ContextCloser,

    Querier,
    Rows,
    Conn,
    Db,
    Adapter,
)

from .columns import (  # noqa
    Column,
    Columns,
)

from .errors import (  # noqa
    Error,

    ColumnError,
    DuplicateColumnNameError,
    MismatchedColumnCountError,

    QueryError,
)

from .funcs import (  # noqa
    query,
    exec,  # noqa
)

from .queries import (  # noqa
    QueryMode,
    Query,
)

from .rows import (  # noqa
    Row,
)



# type: ignore
# ruff: noqa
# flake8: noqa
# Generated from Minisql.g4 by ANTLR 4.13.2
from omlish.antlr._runtime._all import *
if "." in __name__:
    from .MinisqlParser import MinisqlParser
else:
    from MinisqlParser import MinisqlParser

# This class defines a complete listener for a parse tree produced by MinisqlParser.
class MinisqlListener(ParseTreeListener):

    # Enter a parse tree produced by MinisqlParser#singleStmt.
    def enterSingleStmt(self, ctx:MinisqlParser.SingleStmtContext):
        pass

    # Exit a parse tree produced by MinisqlParser#singleStmt.
    def exitSingleStmt(self, ctx:MinisqlParser.SingleStmtContext):
        pass


    # Enter a parse tree produced by MinisqlParser#select.
    def enterSelect(self, ctx:MinisqlParser.SelectContext):
        pass

    # Exit a parse tree produced by MinisqlParser#select.
    def exitSelect(self, ctx:MinisqlParser.SelectContext):
        pass


    # Enter a parse tree produced by MinisqlParser#cteSelect.
    def enterCteSelect(self, ctx:MinisqlParser.CteSelectContext):
        pass

    # Exit a parse tree produced by MinisqlParser#cteSelect.
    def exitCteSelect(self, ctx:MinisqlParser.CteSelectContext):
        pass


    # Enter a parse tree produced by MinisqlParser#cte.
    def enterCte(self, ctx:MinisqlParser.CteContext):
        pass

    # Exit a parse tree produced by MinisqlParser#cte.
    def exitCte(self, ctx:MinisqlParser.CteContext):
        pass


    # Enter a parse tree produced by MinisqlParser#unionSelect.
    def enterUnionSelect(self, ctx:MinisqlParser.UnionSelectContext):
        pass

    # Exit a parse tree produced by MinisqlParser#unionSelect.
    def exitUnionSelect(self, ctx:MinisqlParser.UnionSelectContext):
        pass


    # Enter a parse tree produced by MinisqlParser#unionItem.
    def enterUnionItem(self, ctx:MinisqlParser.UnionItemContext):
        pass

    # Exit a parse tree produced by MinisqlParser#unionItem.
    def exitUnionItem(self, ctx:MinisqlParser.UnionItemContext):
        pass


    # Enter a parse tree produced by MinisqlParser#primarySelect.
    def enterPrimarySelect(self, ctx:MinisqlParser.PrimarySelectContext):
        pass

    # Exit a parse tree produced by MinisqlParser#primarySelect.
    def exitPrimarySelect(self, ctx:MinisqlParser.PrimarySelectContext):
        pass


    # Enter a parse tree produced by MinisqlParser#allSelectItem.
    def enterAllSelectItem(self, ctx:MinisqlParser.AllSelectItemContext):
        pass

    # Exit a parse tree produced by MinisqlParser#allSelectItem.
    def exitAllSelectItem(self, ctx:MinisqlParser.AllSelectItemContext):
        pass


    # Enter a parse tree produced by MinisqlParser#exprSelectItem.
    def enterExprSelectItem(self, ctx:MinisqlParser.ExprSelectItemContext):
        pass

    # Exit a parse tree produced by MinisqlParser#exprSelectItem.
    def exitExprSelectItem(self, ctx:MinisqlParser.ExprSelectItemContext):
        pass


    # Enter a parse tree produced by MinisqlParser#expr.
    def enterExpr(self, ctx:MinisqlParser.ExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#expr.
    def exitExpr(self, ctx:MinisqlParser.ExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#unaryBooleanExpr.
    def enterUnaryBooleanExpr(self, ctx:MinisqlParser.UnaryBooleanExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#unaryBooleanExpr.
    def exitUnaryBooleanExpr(self, ctx:MinisqlParser.UnaryBooleanExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#predicatedBooleanExpr.
    def enterPredicatedBooleanExpr(self, ctx:MinisqlParser.PredicatedBooleanExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#predicatedBooleanExpr.
    def exitPredicatedBooleanExpr(self, ctx:MinisqlParser.PredicatedBooleanExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#binaryBooleanExpr.
    def enterBinaryBooleanExpr(self, ctx:MinisqlParser.BinaryBooleanExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#binaryBooleanExpr.
    def exitBinaryBooleanExpr(self, ctx:MinisqlParser.BinaryBooleanExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#castBooleanExpr.
    def enterCastBooleanExpr(self, ctx:MinisqlParser.CastBooleanExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#castBooleanExpr.
    def exitCastBooleanExpr(self, ctx:MinisqlParser.CastBooleanExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#cmpPredicate.
    def enterCmpPredicate(self, ctx:MinisqlParser.CmpPredicateContext):
        pass

    # Exit a parse tree produced by MinisqlParser#cmpPredicate.
    def exitCmpPredicate(self, ctx:MinisqlParser.CmpPredicateContext):
        pass


    # Enter a parse tree produced by MinisqlParser#isNullPredicate.
    def enterIsNullPredicate(self, ctx:MinisqlParser.IsNullPredicateContext):
        pass

    # Exit a parse tree produced by MinisqlParser#isNullPredicate.
    def exitIsNullPredicate(self, ctx:MinisqlParser.IsNullPredicateContext):
        pass


    # Enter a parse tree produced by MinisqlParser#inListPredicate.
    def enterInListPredicate(self, ctx:MinisqlParser.InListPredicateContext):
        pass

    # Exit a parse tree produced by MinisqlParser#inListPredicate.
    def exitInListPredicate(self, ctx:MinisqlParser.InListPredicateContext):
        pass


    # Enter a parse tree produced by MinisqlParser#inSelectPredicate.
    def enterInSelectPredicate(self, ctx:MinisqlParser.InSelectPredicateContext):
        pass

    # Exit a parse tree produced by MinisqlParser#inSelectPredicate.
    def exitInSelectPredicate(self, ctx:MinisqlParser.InSelectPredicateContext):
        pass


    # Enter a parse tree produced by MinisqlParser#likePredicate.
    def enterLikePredicate(self, ctx:MinisqlParser.LikePredicateContext):
        pass

    # Exit a parse tree produced by MinisqlParser#likePredicate.
    def exitLikePredicate(self, ctx:MinisqlParser.LikePredicateContext):
        pass


    # Enter a parse tree produced by MinisqlParser#arithValueExpr.
    def enterArithValueExpr(self, ctx:MinisqlParser.ArithValueExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#arithValueExpr.
    def exitArithValueExpr(self, ctx:MinisqlParser.ArithValueExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#unaryValueExpr.
    def enterUnaryValueExpr(self, ctx:MinisqlParser.UnaryValueExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#unaryValueExpr.
    def exitUnaryValueExpr(self, ctx:MinisqlParser.UnaryValueExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#primaryValueExpr.
    def enterPrimaryValueExpr(self, ctx:MinisqlParser.PrimaryValueExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#primaryValueExpr.
    def exitPrimaryValueExpr(self, ctx:MinisqlParser.PrimaryValueExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#functionCallExpr.
    def enterFunctionCallExpr(self, ctx:MinisqlParser.FunctionCallExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#functionCallExpr.
    def exitFunctionCallExpr(self, ctx:MinisqlParser.FunctionCallExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#starFunctionCallExpr.
    def enterStarFunctionCallExpr(self, ctx:MinisqlParser.StarFunctionCallExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#starFunctionCallExpr.
    def exitStarFunctionCallExpr(self, ctx:MinisqlParser.StarFunctionCallExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#caseExpr.
    def enterCaseExpr(self, ctx:MinisqlParser.CaseExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#caseExpr.
    def exitCaseExpr(self, ctx:MinisqlParser.CaseExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#selectExpr.
    def enterSelectExpr(self, ctx:MinisqlParser.SelectExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#selectExpr.
    def exitSelectExpr(self, ctx:MinisqlParser.SelectExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#parenExpr.
    def enterParenExpr(self, ctx:MinisqlParser.ParenExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#parenExpr.
    def exitParenExpr(self, ctx:MinisqlParser.ParenExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#simplePrimaryExpr.
    def enterSimplePrimaryExpr(self, ctx:MinisqlParser.SimplePrimaryExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#simplePrimaryExpr.
    def exitSimplePrimaryExpr(self, ctx:MinisqlParser.SimplePrimaryExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#simpleExpr.
    def enterSimpleExpr(self, ctx:MinisqlParser.SimpleExprContext):
        pass

    # Exit a parse tree produced by MinisqlParser#simpleExpr.
    def exitSimpleExpr(self, ctx:MinisqlParser.SimpleExprContext):
        pass


    # Enter a parse tree produced by MinisqlParser#caseItem.
    def enterCaseItem(self, ctx:MinisqlParser.CaseItemContext):
        pass

    # Exit a parse tree produced by MinisqlParser#caseItem.
    def exitCaseItem(self, ctx:MinisqlParser.CaseItemContext):
        pass


    # Enter a parse tree produced by MinisqlParser#over.
    def enterOver(self, ctx:MinisqlParser.OverContext):
        pass

    # Exit a parse tree produced by MinisqlParser#over.
    def exitOver(self, ctx:MinisqlParser.OverContext):
        pass


    # Enter a parse tree produced by MinisqlParser#sortItem.
    def enterSortItem(self, ctx:MinisqlParser.SortItemContext):
        pass

    # Exit a parse tree produced by MinisqlParser#sortItem.
    def exitSortItem(self, ctx:MinisqlParser.SortItemContext):
        pass


    # Enter a parse tree produced by MinisqlParser#aliasedRelation.
    def enterAliasedRelation(self, ctx:MinisqlParser.AliasedRelationContext):
        pass

    # Exit a parse tree produced by MinisqlParser#aliasedRelation.
    def exitAliasedRelation(self, ctx:MinisqlParser.AliasedRelationContext):
        pass


    # Enter a parse tree produced by MinisqlParser#joinRelation.
    def enterJoinRelation(self, ctx:MinisqlParser.JoinRelationContext):
        pass

    # Exit a parse tree produced by MinisqlParser#joinRelation.
    def exitJoinRelation(self, ctx:MinisqlParser.JoinRelationContext):
        pass


    # Enter a parse tree produced by MinisqlParser#selectRelation.
    def enterSelectRelation(self, ctx:MinisqlParser.SelectRelationContext):
        pass

    # Exit a parse tree produced by MinisqlParser#selectRelation.
    def exitSelectRelation(self, ctx:MinisqlParser.SelectRelationContext):
        pass


    # Enter a parse tree produced by MinisqlParser#tableRelation.
    def enterTableRelation(self, ctx:MinisqlParser.TableRelationContext):
        pass

    # Exit a parse tree produced by MinisqlParser#tableRelation.
    def exitTableRelation(self, ctx:MinisqlParser.TableRelationContext):
        pass


    # Enter a parse tree produced by MinisqlParser#parenRelation.
    def enterParenRelation(self, ctx:MinisqlParser.ParenRelationContext):
        pass

    # Exit a parse tree produced by MinisqlParser#parenRelation.
    def exitParenRelation(self, ctx:MinisqlParser.ParenRelationContext):
        pass


    # Enter a parse tree produced by MinisqlParser#groupBy.
    def enterGroupBy(self, ctx:MinisqlParser.GroupByContext):
        pass

    # Exit a parse tree produced by MinisqlParser#groupBy.
    def exitGroupBy(self, ctx:MinisqlParser.GroupByContext):
        pass


    # Enter a parse tree produced by MinisqlParser#qualifiedName.
    def enterQualifiedName(self, ctx:MinisqlParser.QualifiedNameContext):
        pass

    # Exit a parse tree produced by MinisqlParser#qualifiedName.
    def exitQualifiedName(self, ctx:MinisqlParser.QualifiedNameContext):
        pass


    # Enter a parse tree produced by MinisqlParser#identList.
    def enterIdentList(self, ctx:MinisqlParser.IdentListContext):
        pass

    # Exit a parse tree produced by MinisqlParser#identList.
    def exitIdentList(self, ctx:MinisqlParser.IdentListContext):
        pass


    # Enter a parse tree produced by MinisqlParser#ident.
    def enterIdent(self, ctx:MinisqlParser.IdentContext):
        pass

    # Exit a parse tree produced by MinisqlParser#ident.
    def exitIdent(self, ctx:MinisqlParser.IdentContext):
        pass


    # Enter a parse tree produced by MinisqlParser#quotedIdent.
    def enterQuotedIdent(self, ctx:MinisqlParser.QuotedIdentContext):
        pass

    # Exit a parse tree produced by MinisqlParser#quotedIdent.
    def exitQuotedIdent(self, ctx:MinisqlParser.QuotedIdentContext):
        pass


    # Enter a parse tree produced by MinisqlParser#integerNumber.
    def enterIntegerNumber(self, ctx:MinisqlParser.IntegerNumberContext):
        pass

    # Exit a parse tree produced by MinisqlParser#integerNumber.
    def exitIntegerNumber(self, ctx:MinisqlParser.IntegerNumberContext):
        pass


    # Enter a parse tree produced by MinisqlParser#decimalNumber.
    def enterDecimalNumber(self, ctx:MinisqlParser.DecimalNumberContext):
        pass

    # Exit a parse tree produced by MinisqlParser#decimalNumber.
    def exitDecimalNumber(self, ctx:MinisqlParser.DecimalNumberContext):
        pass


    # Enter a parse tree produced by MinisqlParser#floatNumber.
    def enterFloatNumber(self, ctx:MinisqlParser.FloatNumberContext):
        pass

    # Exit a parse tree produced by MinisqlParser#floatNumber.
    def exitFloatNumber(self, ctx:MinisqlParser.FloatNumberContext):
        pass


    # Enter a parse tree produced by MinisqlParser#string.
    def enterString(self, ctx:MinisqlParser.StringContext):
        pass

    # Exit a parse tree produced by MinisqlParser#string.
    def exitString(self, ctx:MinisqlParser.StringContext):
        pass


    # Enter a parse tree produced by MinisqlParser#null.
    def enterNull(self, ctx:MinisqlParser.NullContext):
        pass

    # Exit a parse tree produced by MinisqlParser#null.
    def exitNull(self, ctx:MinisqlParser.NullContext):
        pass


    # Enter a parse tree produced by MinisqlParser#true.
    def enterTrue(self, ctx:MinisqlParser.TrueContext):
        pass

    # Exit a parse tree produced by MinisqlParser#true.
    def exitTrue(self, ctx:MinisqlParser.TrueContext):
        pass


    # Enter a parse tree produced by MinisqlParser#false.
    def enterFalse(self, ctx:MinisqlParser.FalseContext):
        pass

    # Exit a parse tree produced by MinisqlParser#false.
    def exitFalse(self, ctx:MinisqlParser.FalseContext):
        pass


    # Enter a parse tree produced by MinisqlParser#setQuantifier.
    def enterSetQuantifier(self, ctx:MinisqlParser.SetQuantifierContext):
        pass

    # Exit a parse tree produced by MinisqlParser#setQuantifier.
    def exitSetQuantifier(self, ctx:MinisqlParser.SetQuantifierContext):
        pass


    # Enter a parse tree produced by MinisqlParser#joinType.
    def enterJoinType(self, ctx:MinisqlParser.JoinTypeContext):
        pass

    # Exit a parse tree produced by MinisqlParser#joinType.
    def exitJoinType(self, ctx:MinisqlParser.JoinTypeContext):
        pass


    # Enter a parse tree produced by MinisqlParser#cmpOp.
    def enterCmpOp(self, ctx:MinisqlParser.CmpOpContext):
        pass

    # Exit a parse tree produced by MinisqlParser#cmpOp.
    def exitCmpOp(self, ctx:MinisqlParser.CmpOpContext):
        pass


    # Enter a parse tree produced by MinisqlParser#arithOp.
    def enterArithOp(self, ctx:MinisqlParser.ArithOpContext):
        pass

    # Exit a parse tree produced by MinisqlParser#arithOp.
    def exitArithOp(self, ctx:MinisqlParser.ArithOpContext):
        pass


    # Enter a parse tree produced by MinisqlParser#unaryOp.
    def enterUnaryOp(self, ctx:MinisqlParser.UnaryOpContext):
        pass

    # Exit a parse tree produced by MinisqlParser#unaryOp.
    def exitUnaryOp(self, ctx:MinisqlParser.UnaryOpContext):
        pass


    # Enter a parse tree produced by MinisqlParser#unquotedIdent.
    def enterUnquotedIdent(self, ctx:MinisqlParser.UnquotedIdentContext):
        pass

    # Exit a parse tree produced by MinisqlParser#unquotedIdent.
    def exitUnquotedIdent(self, ctx:MinisqlParser.UnquotedIdentContext):
        pass



del MinisqlParser


import abc

from ..processing.base import Processor
from ..processing.priority import ProcessorPriority
from ..processing.registry import register_processor_type


##


@register_processor_type(priority=ProcessorPriority.POST_SLOTS)
class UpdateAbstractMethodsProcessor(Processor):
    def process(self, cls: type) -> type:
        abc.update_abstractmethods(cls)  # noqa
        return cls



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .accessor import TypedValuesAccessor
from .consumer import TypedValuesConsumer
from .values import TypedValue
from .values import UniqueTypedValue


TypedValueT = ta.TypeVar('TypedValueT', bound='TypedValue')

UniqueTypedValueT = ta.TypeVar('UniqueTypedValueT', bound='UniqueTypedValue')


##


@dc.dataclass()
class DuplicateUniqueTypedValueError(Exception, ta.Generic[UniqueTypedValueT]):
    cls: type[UniqueTypedValueT]
    new: UniqueTypedValueT
    old: UniqueTypedValueT


class TypedValues(
    TypedValuesAccessor[TypedValueT],
    lang.Final,
    ta.Generic[TypedValueT],
):
    def __init__(
            self,
            *tvs: TypedValueT,
            override: bool = False,
            check_type: type | tuple[type, ...] | None = None,
    ) -> None:
        if hasattr(self, '_tup'):
            # When __new__ returns the empty singleton __init__ will still be called.
            if self is not self._EMPTY:
                raise RuntimeError
            return

        super().__init__()

        tmp: list = []
        udct: dict = {}
        for tv in tvs:
            if check_type is not None:
                check.isinstance(tv, check_type)
            if isinstance(tv, UniqueTypedValue):
                utvc = tv._unique_typed_value_cls  # noqa
                if not override:
                    try:
                        exu = udct[utvc]
                    except KeyError:
                        pass
                    else:
                        raise DuplicateUniqueTypedValueError(utvc, tv, check.single(exu))
                ulst = udct.setdefault(utvc, [])
                ulst.append(tv)
                tmp.append((utvc, tv, ulst, len(ulst)))
            elif isinstance(tv, TypedValue):
                tmp.append(tv)
            else:
                raise TypeError(tv)

        lst: list = []
        dct: dict = {}
        for obj in tmp:
            if isinstance(obj, tuple):
                utvc, tv, ulst, idx = obj
                if idx == len(ulst):
                    lst.append(tv)
                    dct[utvc] = tv
            else:
                tv = obj
                lst.append(tv)
                dct.setdefault(type(tv), []).append(tv)

        self._tup: tuple[TypedValueT, ...] = tuple(lst)
        self._dct: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]] = {
            k: tuple(v) if isinstance(v, list) else v
            for k, v in dct.items()
        }
        self._dct2: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]] = {
            **self._dct,
            **{type(v): v for v in self._dct.values() if isinstance(v, UniqueTypedValue)},  # type: ignore[misc]
        }

    #

    _EMPTY: ta.ClassVar['TypedValues']

    @classmethod
    def empty(cls) -> 'TypedValues':
        return cls._EMPTY

    def __new__(cls, *tvs, **kwargs):  # noqa
        if not tvs:
            try:
                return cls._EMPTY
            except AttributeError:
                pass
        return super().__new__(cls)

    #

    def without(self, *tys: type) -> ta.Iterator[TypedValueT]:
        for o in self._tup:
            if isinstance(o, tys):
                continue
            yield o

    #

    def update(self, *tvs, override: bool = False) -> 'TypedValues':
        return TypedValues(*self._tup, *tvs, override=override)

    def discard(self, *tys: type) -> 'TypedValues':
        return TypedValues(*self.without(*tys))

    #

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(map(repr, self._tup))})'

    #

    _set: frozenset[TypedValueT]

    def to_set(self) -> frozenset[TypedValueT]:
        try:
            return self._set
        except AttributeError:
            pass
        s = frozenset(self._tup)
        self._set = s
        return s

    #

    _hash: int

    def __hash__(self) -> int:
        try:
            return self._hash
        except AttributeError:
            pass
        h = hash(self._tup)
        self._hash = h
        return h

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            other is self or
            other._tup == self._tup
        )

    #

    def __iter__(self) -> ta.Iterator[TypedValueT]:
        return iter(self._tup)

    def __len__(self) -> int:
        return len(self._tup)

    def __bool__(self) -> bool:
        return bool(self._tup)

    #

    def keys(self) -> ta.KeysView[type[TypedValueT]]:
        return self._dct.keys()

    def values(self) -> ta.ValuesView[TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.values()

    def items(self) -> ta.ItemsView[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.items()

    #

    def consume(self) -> TypedValuesConsumer[TypedValueT]:
        return TypedValuesConsumer(self._dct.items())

    #

    def _typed_value_contains(self, cls):
        return cls in self._dct2

    def _typed_value_getitem(self, key):
        if isinstance(key, int):
            return self._tup[key]
        elif isinstance(key, type):
            return self._dct2[check.issubclass(key, TypedValue)]
        else:
            raise TypeError(key)

    _any_dct: dict[type, tuple[TypedValueT, ...]]

    def _typed_value_get_any(self, cls):
        try:
            any_dct = self._any_dct
        except AttributeError:
            any_dct = {}
            self._any_dct = any_dct

        try:
            return any_dct[cls]
        except KeyError:
            pass

        ret = tuple(tv for tv in self if isinstance(tv, cls))
        any_dct[cls] = ret
        return ret


TypedValues._EMPTY = TypedValues()  # noqa



import datetime


##


def utcnow() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.UTC)


def utcfromtimestamp(ts: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(ts, tz=datetime.UTC)



import abc
import functools
import typing as ta


T = ta.TypeVar('T')


##


class Maybe(ta.Generic[T]):
    @property
    @abc.abstractmethod
    def present(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def must(self) -> T:
        raise NotImplementedError

    @classmethod
    def just(cls, v: T) -> 'Maybe[T]':
        return tuple.__new__(_Maybe, (v,))  # noqa

    _empty: ta.ClassVar['Maybe']

    @classmethod
    def empty(cls) -> 'Maybe[T]':
        return Maybe._empty


class _Maybe(Maybe[T], tuple):
    __slots__ = ()

    def __init_subclass__(cls, **kwargs):
        raise TypeError

    @property
    def present(self) -> bool:
        return bool(self)

    def must(self) -> T:
        if not self:
            raise ValueError
        return self[0]


Maybe._empty = tuple.__new__(_Maybe, ())  # noqa


##


@functools.singledispatch
def as_maybe(obj: ta.Any) -> Maybe:
    raise TypeError(obj)


@as_maybe.register
def _(obj: Maybe) -> Maybe:
    return obj



# ruff: noqa: UP006 UP007
# @omlish-lite
import configparser
import io
import typing as ta


IniSectionSettingsMap = ta.Mapping[str, ta.Mapping[str, ta.Union[str, ta.Sequence[str]]]]  # ta.TypeAlias


##


def extract_ini_sections(cp: configparser.ConfigParser) -> IniSectionSettingsMap:
    config_dct: ta.Dict[str, ta.Any] = {}
    for sec in cp.sections():
        cd = config_dct
        for k in sec.split('.'):
            cd = cd.setdefault(k, {})
        cd.update(cp.items(sec))
    return config_dct


##


def render_ini_sections(
        settings_by_section: IniSectionSettingsMap,
) -> str:
    out = io.StringIO()

    for i, (section, settings) in enumerate(settings_by_section.items()):
        if i:
            out.write('\n')

        out.write(f'[{section}]\n')

        for k, v in settings.items():
            if isinstance(v, str):
                out.write(f'{k}={v}\n')
            else:
                for vv in v:
                    out.write(f'{k}={vv}\n')

    return out.getvalue()



"""
TODO:
 - option to coordinate with objects and omit if empty / render unboxed
"""
import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactoryMatchClass
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactoryMatchClass
from ..values import Value


##


@dc.dataclass(frozen=True)
class MaybeMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        m: lang.Maybe = o
        if m.present:
            return [self.e.marshal(ctx, m.must())]
        else:
            return []


class MaybeMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is lang.Maybe)
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Generic)
        return MaybeMarshaler(ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, lang.Maybe))
    def _build_concrete(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return MaybeMarshaler(ctx.make(ta.Any))


@dc.dataclass(frozen=True)
class MaybeUnmarshaler(Unmarshaler):
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if v:
            return lang.just(self.e.unmarshal(ctx, check.single(v)))  # type: ignore
        else:
            return lang.empty()


class MaybeUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is lang.Maybe)
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Generic)
        return MaybeUnmarshaler(ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, lang.Maybe))
    def _build_concrete(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return MaybeUnmarshaler(ctx.make(ta.Any))



import typing as ta

from .. import check
from .. import reflect as rfl
from .values import TypedValue


##


def reflect_typed_values_impls(rty: rfl.Type) -> set[type[TypedValue]]:
    tv_cls_set: set[type[TypedValue]] = set()

    todo = [rty]
    seen = set()
    while todo:
        cur = todo.pop()
        if cur in seen:
            continue
        seen.add(cur)

        if isinstance(cur, rfl.Union):
            todo.extend(cur.args)
        elif isinstance(cur, ta.TypeVar):
            todo.append(rfl.type_(rfl.get_type_var_bound(cur)))
        else:
            tv_cls_set.add(check.issubclass(check.isinstance(cur, type), TypedValue))

    return tv_cls_set



import typing as ta

from ... import dataclasses as dc
from ... import lang
from .base import Node
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .idents import CanIdent
from .idents import Ident
from .relations import CanRelation
from .relations import Relation
from .relations import RelationBuilder
from .selects import Select
from .stmts import Stmt


##


class Values(Node, lang.Final):
    vs: ta.Sequence[Expr] = dc.xfield(coerce=tuple)


class Insert(Stmt, lang.Final):
    columns: ta.Sequence[Ident] = dc.xfield(coerce=tuple)
    into: Relation
    data: Values | Select


CanValues: ta.TypeAlias = Values | ta.Sequence[CanExpr]


class InsertBuilder(RelationBuilder, ExprBuilder):
    def values(self, vs: CanValues) -> Values:
        if isinstance(vs, Values):
            return vs
        else:
            return Values(tuple(self.expr(v) for v in vs))

    def insert(
            self,
            columns: ta.Sequence[CanIdent],
            into: CanRelation,
            data: Select | Values | ta.Sequence[CanExpr],
    ) -> Insert:
        return Insert(
            columns=tuple(self.ident(c) for c in columns),
            into=self.relation(into),
            data=data if isinstance(data, Select) else self.values(data),
        )



import collections.abc
import io
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import dispatch


T = ta.TypeVar('T')


Part: ta.TypeAlias = ta.Union[str, ta.Sequence['Part'], 'DataPart']
PartT = ta.TypeVar('PartT', bound=Part)


##


def check_part(o: PartT) -> PartT:
    if isinstance(o, (str, DataPart)):
        pass
    elif isinstance(o, ta.Sequence):
        for c in o:
            check_part(c)
    else:
        raise TypeError(o)
    return o


def check_opt_part(o: PartT | None) -> PartT | None:
    if o is None:
        return None
    return check_part(o)


##


class DataPart(dc.Case):
    pass


class Wrap(DataPart):
    part: Part
    wrapper: tuple[str, str] = ('(', ')')


class List(DataPart):
    parts: ta.Sequence[Part | None]
    delimiter: str = ','  # FIXME: Part
    trailer: bool = False


class Concat(DataPart):
    parts: ta.Sequence[Part]


class Block(DataPart):
    parts: ta.Sequence[Part]


class Section(DataPart):
    parts: ta.Sequence[Part]


class Meta(DataPart):
    node: ta.Any


##


class PartTransform:
    def __call__(self, part: Part | None) -> Part:
        return self._transform(part)

    @dispatch.method
    def _transform(self, part: Part | None) -> Part:
        raise TypeError(part)

    @_transform.register
    def _transform_none(self, part: None) -> Part:
        return []

    @_transform.register
    def _transform_str(self, part: str) -> Part:
        return part

    @_transform.register
    def _transform_sequence(self, part: collections.abc.Sequence) -> Part:
        return [self(c) for c in part]

    @_transform.register
    def _transform_wrap(self, part: Wrap) -> Part:
        return Wrap(self(part.part), part.wrapper)

    @_transform.register
    def _transform_list(self, part: List) -> Part:
        return List([self(c) for c in part.parts], part.delimiter, part.trailer)

    @_transform.register
    def _transform_concat(self, part: Concat) -> Part:
        return Concat([self(c) for c in part.parts])

    @_transform.register
    def _transform_block(self, part: Block) -> Part:
        return Block([self(c) for c in part.parts])

    @_transform.register
    def _transform_section(self, part: Section) -> Part:
        return Section([self(c) for c in part.parts])

    @_transform.register
    def _transform_meta(self, part: Meta) -> Meta:
        return part


##


class RemoveMetas(PartTransform):
    @PartTransform._transform.register  # noqa
    def _transform_meta(self, part: Meta) -> Part:
        return []


remove_metas = RemoveMetas()


##


def _drop_empties(it: ta.Iterable[T]) -> list[T]:
    return [
        e  # type: ignore
        for e in it
        if not (
            isinstance(e, collections.abc.Sequence) and
            not e and
            not isinstance(e, str)
        )
    ]


class CompactPart(PartTransform):
    @PartTransform._transform.register  # noqa
    def _transform_sequence(self, part: collections.abc.Sequence) -> Part:
        return _drop_empties(self(c) for c in part)

    @PartTransform._transform.register  # noqa
    def _transform_list(self, part: List) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return List(parts, part.delimiter, part.trailer) if parts else []

    @PartTransform._transform.register  # noqa
    def _transform_concat(self, part: Concat) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return Concat(parts) if parts else []

    @PartTransform._transform.register  # noqa
    def _transform_block(self, part: Block) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return Block(parts) if parts else []

    @PartTransform._transform.register  # noqa
    def _transform_section(self, part: Section) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return Section(parts) if parts else []


compact_part = CompactPart()


##


class PartRenderer:
    def __init__(
            self,
            buf: io.StringIO,
            *,
            indent: int | str = 4,
    ) -> None:
        super().__init__()

        self._buf = buf

        self._indents = 0
        if isinstance(indent, int):
            indent = ' ' * indent
        self._indent = indent

        self._blank_lines = 0
        self._has_indented = False

    def _write(self, s: str) -> None:
        check.not_in('\n', s)
        if not self._has_indented:
            self._buf.write(self._indent * self._indents)
            self._blank_lines = 0
            self._has_indented = True

        self._buf.write(s)

    def _write_newline(self, n: int = 1) -> None:
        check.arg(n >= 0)
        d = n - self._blank_lines
        if d > 0:
            self._buf.write('\n' * n)
            self._blank_lines += n
            self._has_indented = False

    def __call__(self, part: Part | None) -> None:
        return self._render(part)

    @dispatch.method
    def _render(self, part: Part | None) -> None:
        raise TypeError(part)

    @_render.register
    def _render_none(self, part: None) -> None:
        pass

    @_render.register
    def _render_str(self, part: str) -> None:
        self._write(part)

    @_render.register
    def _render_sequence(self, part: collections.abc.Sequence) -> None:
        for i, c in enumerate(part):
            if i:
                self._write(' ')
            self(c)

    @_render.register
    def _render_wrap(self, part: Wrap) -> None:
        self._write(part.wrapper[0])
        self(part.part)
        self._write(part.wrapper[1])

    @_render.register
    def _render_list(self, part: List) -> None:
        for i, c in enumerate(part.parts):
            if i:
                self._write(part.delimiter + ' ')
            self(c)
        if part.trailer:
            self._write(part.delimiter)

    @_render.register
    def _render_concat(self, part: Concat) -> None:
        for c in part.parts:
            self(c)

    @_render.register
    def _render_block(self, part: Block) -> None:
        for c in part.parts:
            self(c)
            self._write_newline()

    @_render.register
    def _render_section(self, part: Section) -> None:
        self._indents += 1
        try:
            for c in part.parts:
                self(c)
        finally:
            self._indents -= 1


def render_part(part: Part, buf: io.StringIO | None = None) -> io.StringIO:
    if buf is None:
        buf = io.StringIO()
    PartRenderer(buf)(part)
    return buf


def render(part: Part) -> str:
    part = remove_metas(part)
    part = compact_part(part)
    return render_part(part).getvalue()



"""
TODO:
 - sync vs async
"""
import abc
import typing as ta

from ... import lang
from .columns import Column
from .columns import Columns
from .queries import Query
from .rows import Row


##


class Closer(lang.Abstract):
    def close(self) -> None:
        pass


class ContextCloser(Closer):
    def __enter__(self) -> ta.Self:
        return self

    @ta.final
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


##


class Querier(ContextCloser, lang.Abstract):
    @property
    @abc.abstractmethod
    def adapter(self) -> 'Adapter':
        raise NotImplementedError

    @abc.abstractmethod
    def query(self, query: Query) -> 'Rows':  # ta.Raises[QueryError]
        raise NotImplementedError


##


class Rows(ContextCloser, lang.Abstract):
    @property
    @abc.abstractmethod
    def columns(self) -> Columns:
        raise NotImplementedError

    @ta.final
    def __iter__(self) -> ta.Self:
        return self

    @abc.abstractmethod
    def __next__(self) -> Row:  # ta.Raises[StopIteration]
        raise NotImplementedError


##


class Conn(Querier, lang.Abstract):  # noqa
    pass


##


class Db(Querier, lang.Abstract):  # noqa
    @abc.abstractmethod
    def connect(self) -> Conn:
        raise NotImplementedError

    def query(self, query: Query) -> Rows:
        with self.connect() as conn:
            return conn.query(query)


##


class Adapter(lang.Abstract):
    @abc.abstractmethod
    def scan_type(self, c: Column) -> type:
        raise NotImplementedError



import abc
import typing as ta

from .. import dataclasses as dc
from .. import lang


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class Origin:
    lst: ta.Sequence[str]


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class Origins:
    lst: ta.Sequence[Origin]

    def __iter__(self) -> ta.Iterator[Origin]:
        yield from self.lst


class HasOrigins(lang.Abstract):
    @property
    @abc.abstractmethod
    def origins(self) -> Origins:
        raise NotImplementedError



"""
MIT License

Copyright (c) 2022 Gabriel Ochsenhofer

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
# ruff: noqa: SLF001
import typing as ta


T = ta.TypeVar('T')

Comparer: ta.TypeAlias = ta.Callable[[T, T], int]


##


class TreapNode(ta.Generic[T]):
    __slots__ = ('_value', '_priority', '_left', '_right')

    def __init__(
            self,
            *,
            _value: T,
            _priority: int,
            _left: ta.Optional['TreapNode[T]'],
            _right: ta.Optional['TreapNode[T]'],
    ) -> None:
        super().__init__()

        self._value = _value
        self._priority = _priority
        self._left = _left
        self._right = _right

    def __repr__(self) -> str:
        return f'TreapNode(value={self._value!r}, priority={self._priority!r})'

    @property
    def value(self) -> T:
        return self._value

    @property
    def priority(self) -> int:
        return self._priority

    @property
    def left(self) -> ta.Optional['TreapNode[T]']:
        return self._left

    @property
    def right(self) -> ta.Optional['TreapNode[T]']:
        return self._right

    def __iter__(self) -> ta.Iterator[T]:
        if self._left is not None:
            yield from self._left
        yield self._value
        if self._right is not None:
            yield from self._right


def find(n: TreapNode[T] | None, v: T, c: Comparer[T]) -> TreapNode[T] | None:
    while True:
        if n is None:
            return None
        diff = c(n._value, v)  # noqa
        if diff == 0:
            return n
        elif diff < 0:
            n = n._right  # noqa
        else:
            n = n._left  # noqa


def place(n: TreapNode[T] | None, v: T, c: Comparer[T], *, desc: bool = False) -> list[TreapNode[T]]:
    ret: list[TreapNode[T]] = []
    while True:
        if n is None:
            break
        diff = c(n._value, v)  # noqa
        if diff == 0:
            ret.append(n)
            break
        elif diff < 0:
            if desc:
                ret.append(n)
            n = n._right  # noqa
        else:
            if not desc:
                ret.append(n)
            n = n._left  # noqa
    return ret


def union(
        n: TreapNode[T] | None,
        other: TreapNode[T] | None,
        c: Comparer[T],
        overwrite: bool,
) -> TreapNode[T] | None:
    if n is None:
        return other
    if other is None:
        return n
    if n._priority < other._priority:
        other, n, overwrite = n, other, not overwrite
    left, dupe, right = split(other, n._value, c)
    value = n._value
    if overwrite and dupe is not None:
        value = dupe._value
    left = union(n._left, left, c, overwrite)
    right = union(n._right, right, c, overwrite)
    return TreapNode(_value=value, _priority=n._priority, _left=left, _right=right)


def split(
        n: TreapNode[T] | None,
        v: T,
        c: Comparer[T],
) -> tuple[
    TreapNode[T] | None,
    TreapNode[T] | None,
    TreapNode[T] | None,
]:
    tmp: TreapNode[T] = TreapNode(_value=None, _priority=0, _left=None, _right=None)  # type: ignore
    leftp, rightp = [tmp, 'l'], [tmp, 'r']

    def setp(p, o):
        t, s = p
        if s == 'l':
            t._left = o
        elif s == 'r':
            t._right = o
        else:
            raise ValueError(p)

    cur: TreapNode[T] | None = n
    while True:
        if cur is None:
            setp(leftp, None)
            setp(rightp, None)
            return tmp._left, None, tmp._right  # noqa

        d = c(cur._value, v)
        if d < 0:
            root = TreapNode(_value=cur._value, _priority=cur._priority, _left=cur._left, _right=None)
            setp(leftp, root)
            leftp[0], leftp[1] = root, 'r'
            cur = cur._right
        elif d > 0:
            root = TreapNode(_value=cur._value, _priority=cur._priority, _left=None, _right=cur._right)
            setp(rightp, root)
            rightp[0], rightp[1] = root, 'l'
            cur = cur._left
        else:
            root = TreapNode(_value=cur._value, _priority=cur._priority, _left=None, _right=None)
            setp(leftp, cur._left)
            setp(rightp, cur._right)
            return tmp._left, root, tmp._right


def intersect(
        n: TreapNode[T] | None,
        other: TreapNode[T] | None,
        c: Comparer[T],
) -> TreapNode[T] | None:
    if n is None or other is None:
        return None

    if n._priority < other._priority:
        n, other = other, n

    left, found, right = split(other, n._value, c)
    left = intersect(n._left, left, c)
    right = intersect(n._right, right, c)

    if found is None:
        # TODO: use a destructive join as both left/right are copies
        return _join(left, right)

    return TreapNode(_value=n._value, _priority=n._priority, _left=left, _right=right)


def delete(n: TreapNode[T] | None, v: T, c: Comparer[T]) -> TreapNode[T] | None:
    left, _, right = split(n, v, c)
    return _join(left, right)


def diff(n: TreapNode[T] | None, other: TreapNode[T] | None, c: Comparer[T]) -> TreapNode[T] | None:
    if n is None or other is None:
        return n

    # TODO -- use count
    if n._priority >= other._priority:
        left, dupe, right = split(other, n._value, c)
        left, right = diff(n._left, left, c), diff(n._right, right, c)
        if dupe is not None:
            return _join(left, right)
        return TreapNode(_value=n._value, _priority=n._priority, _left=left, _right=right)

    left, _, right = split(n, other._value, c)
    left = diff(left, other._left, c)
    right = diff(right, other._right, c)
    return _join(left, right)


def _join(n: TreapNode[T] | None, other: TreapNode[T] | None) -> TreapNode[T] | None:
    result: TreapNode[T] | None = None
    resultp: list[ta.Any] = [None, None]

    def setresultp(o):
        t, s = resultp
        if t is None:
            nonlocal result
            result = o
        elif s == 'l':
            t._left = o
        elif s == 'r':
            t._right = o
        else:
            raise ValueError(resultp)

    cur: TreapNode[T] | None = n
    while True:
        if cur is None:
            setresultp(other)
            return result

        if other is None:
            setresultp(cur)
            return result

        if cur._priority <= other._priority:
            root = TreapNode(_value=cur._value, _priority=cur._priority, _left=cur._left, _right=None)
            setresultp(root)
            resultp[0], resultp[1] = root, 'r'
            cur = cur._right
        else:
            root = TreapNode(_value=other._value, _priority=other._priority, _left=None, _right=other._right)
            setresultp(root)
            resultp[0], resultp[1] = root, 'l'
            other = other._left



"""
FIXME:
 - macos pipe size lol, and just like checking at all
"""
import contextlib
import fcntl
import os
import tempfile
import typing as ta


##


class SubprocessFileInput(ta.NamedTuple):
    file_path: str
    pass_fds: ta.Sequence[int]


SubprocessFileInputMethod: ta.TypeAlias = ta.Callable[[bytes], ta.ContextManager[SubprocessFileInput]]


@contextlib.contextmanager
def temp_subprocess_file_input(buf: bytes) -> ta.Iterator[SubprocessFileInput]:
    with tempfile.NamedTemporaryFile(delete=True) as kf:
        kf.write(buf)
        kf.flush()
        yield SubprocessFileInput(kf.name, [])


@contextlib.contextmanager
def pipe_fd_subprocess_file_input(buf: bytes) -> ta.Iterator[SubprocessFileInput]:
    rfd, wfd = os.pipe()
    closed_wfd = False
    try:
        if hasattr(fcntl, 'F_SETPIPE_SZ'):
            fcntl.fcntl(wfd, fcntl.F_SETPIPE_SZ, max(len(buf), 0x1000))
        n = os.write(wfd, buf)
        if n != len(buf):
            raise OSError(f'Failed to write data to pipe: {n=} {len(buf)=}')
        os.close(wfd)
        closed_wfd = True
        yield SubprocessFileInput(f'/dev/fd/{rfd}', [rfd])
    finally:
        if not closed_wfd:
            os.close(wfd)
        os.close(rfd)



"""
TODO:
 - !!! lighter weight bound methods
  - keymaker overhead less important than not rebuilding a whole dc every __get__ on a new instance
 - !! specialize nullary, explicit kwarg
 - !! use c-backed functools.cache if possible
  - also just riic
 - !! reconcile A().f() with A.f(A())
  - unbound descriptor *should* still hit instance cache
 - integrate / expose with collections.cache
 - weakrefs (selectable by arg)
 - more rigorous descriptor pickling
  - must support free functions (which have no instance nor owner)
  - 'staticmethod' or effective equiv - which must resolve to the shared instance
   - and must be transient?
 - use __transient_dict__ to support common state nuking
 - use __set_name__ ?
"""
import dataclasses as dc
import functools
import inspect
import typing as ta

from ..classes.abstract import Abstract
from ..contextmanagers import DefaultLockable
from ..contextmanagers import default_lock
from ..descriptors import unwrap_func
from ..descriptors import unwrap_func_with_partials
from ..params import KwargsParam
from ..params import Param
from ..params import ParamSeparator
from ..params import ParamSpec
from ..params import ValParam
from ..params import param_render


P = ta.ParamSpec('P')
T = ta.TypeVar('T')
CallableT = ta.TypeVar('CallableT', bound=ta.Callable)
CacheKeyMaker: ta.TypeAlias = ta.Callable[..., tuple]


##


def _nullary_cache_key_maker():
    return ()


def _self_cache_key_maker(self):
    return (self,)


_PRE_MADE_CACHE_KEY_MAKERS = [
    _nullary_cache_key_maker,
    _self_cache_key_maker,
]


_PRE_MADE_CACHE_KEY_MAKERS_BY_PARAM_SPEC: ta.Mapping[ParamSpec, CacheKeyMaker] = {
    ParamSpec.inspect(fn): fn  # type: ignore
    for fn in _PRE_MADE_CACHE_KEY_MAKERS
}


##


def _make_unwrapped_cache_key_maker(
        fn: ta.Callable,
        *,
        bound: bool = False,
) -> CacheKeyMaker:
    if inspect.isgeneratorfunction(fn) or inspect.iscoroutinefunction(fn):
        raise TypeError(fn)

    ps = ParamSpec.inspect(
        fn,
        offset=1 if bound else 0,
        strip_annotations=True,
    )

    if not len(ps):
        return _nullary_cache_key_maker

    if not ps.has_defaults:
        try:
            return _PRE_MADE_CACHE_KEY_MAKERS_BY_PARAM_SPEC[ps]
        except KeyError:
            pass

    builtin_pfx = '__cache_key_maker__'
    ns = {
        (builtin_tuple := builtin_pfx + 'tuple'): tuple,
        (builtin_sorted := builtin_pfx + 'sorted'): sorted,
    }

    src_params = []
    src_vals = []
    kwargs_name = None
    for p in ps:
        if isinstance(p, ParamSeparator):
            src_params.append(p.value)
            continue

        if not isinstance(p, Param):
            raise TypeError(p)

        if isinstance(p, ValParam) and p.default.present:
            ns[p.name] = p.default

        src_params.append(param_render(
            p,
            render_default=lambda _: p.name,  # noqa
        ))

        if isinstance(p, KwargsParam):
            kwargs_name = p.name
        else:
            src_vals.append(p.name)

    if kwargs_name is not None:
        src_vals.append(f'{builtin_tuple}({builtin_sorted}({kwargs_name}.items()))')

    rendered = (
        f'def __func__({", ".join(src_params)}):\n'
        f'    return ({", ".join(src_vals)}{"," if len(src_vals) == 1 else ""})\n'
    )
    exec(rendered, ns)

    kfn: CacheKeyMaker = ns['__func__']  # type: ignore[assignment]
    return kfn


def _make_cache_key_maker(
        fn: ta.Any,
        *,
        bound: bool = False,
):
    fn, partials = unwrap_func_with_partials(fn)

    kfn = _make_unwrapped_cache_key_maker(fn, bound=bound)

    for part in partials[::-1]:
        kfn = functools.partial(kfn, *part.args, **part.keywords)

    return kfn


##


class _CachedFunction(ta.Generic[T], Abstract):
    @dc.dataclass(frozen=True, kw_only=True)
    class Opts:
        map_maker: ta.Callable[[], ta.MutableMapping] = dict
        lock: DefaultLockable = None
        transient: bool = False
        no_wrapper_update: bool = False

    def __init__(
            self,
            fn: ta.Callable[P, T],
            *,
            opts: Opts = Opts(),
            key_maker: ta.Callable[..., tuple] | None = None,
            values: ta.MutableMapping | None = None,
            value_fn: ta.Callable[P, T] | None = None,
    ) -> None:
        super().__init__()

        self._fn = (fn,)
        self._opts = opts
        self._key_maker = key_maker if key_maker is not None else _make_cache_key_maker(fn)

        self._lock = default_lock(opts.lock, False)() if opts.lock is not None else None
        self._values = values if values is not None else opts.map_maker()
        self._value_fn = value_fn if value_fn is not None else fn
        if not self._opts.no_wrapper_update:
            functools.update_wrapper(self, fn)

    @property
    def _fn(self):
        return self.__fn

    @_fn.setter
    def _fn(self, x):
        self.__fn = x

    def reset(self) -> None:
        self._values = {}

    def __bool__(self) -> bool:
        raise TypeError

    def __call__(self, *args, **kwargs) -> T:
        k = self._key_maker(*args, **kwargs)

        try:
            return self._values[k]
        except KeyError:
            pass

        if self._lock is not None:
            with self._lock:
                try:
                    return self._values[k]
                except KeyError:
                    pass

                value = self._value_fn(*args, **kwargs)

        else:
            value = self._value_fn(*args, **kwargs)

        self._values[k] = value
        return value


#


class _FreeCachedFunction(_CachedFunction[T]):
    @classmethod
    def _unpickle(
            cls,
            fn,
            opts,
            values,
    ):
        return cls(
            fn,
            opts=opts,
            values=values,
        )

    def __reduce__(self):
        return (
            _FreeCachedFunction._unpickle,
            (
                self._fn,
                self._opts,
                self._values if not self._opts.transient else None,
            ),
        )


#


class _DescriptorCachedFunction(_CachedFunction[T]):
    def __init__(
            self,
            fn: ta.Callable[P, T],
            scope: ta.Any,  # classmethod | None
            *,
            instance: ta.Any = None,
            owner: ta.Any = None,
            name: str | None = None,
            **kwargs,
    ) -> None:
        super().__init__(fn, **kwargs)

        self._scope = scope
        self._instance = instance
        self._owner = owner
        self._name = name if name is not None else unwrap_func(fn).__name__
        self._bound_key_maker = None

    @classmethod
    def _unpickle(
            cls,
            scope,
            instance,
            owner,
            name,
            values,
    ):
        if scope is not None:
            raise NotImplementedError

        if instance is None:
            raise RuntimeError
        obj = type(instance)

        desc: _DescriptorCachedFunction = object.__getattribute__(obj, name)
        if not isinstance(desc, cls):
            raise TypeError(desc)
        if (desc._instance is not None or desc._owner is not None):
            raise RuntimeError

        return desc._bind(
            instance,
            owner,
            values=values,
        )

    def __reduce__(self):
        if self._scope is not None:
            raise NotImplementedError

        if not (self._instance is not None or self._owner is not None):
            raise RuntimeError

        return (
            _DescriptorCachedFunction._unpickle,
            (
                self._scope,
                self._instance,
                self._owner,
                self._name,
                self._values if not self._opts.transient else None,
            ),
        )

    def _bind(
            self,
            instance,
            owner=None,
            *,
            values: ta.MutableMapping | None = None,
    ):
        scope = self._scope
        if owner is self._owner and (instance is self._instance or scope is classmethod):
            return self

        fn, = self._fn
        name = self._name
        bound_fn = fn.__get__(instance, owner)
        if self._bound_key_maker is None:
            self._bound_key_maker = _make_cache_key_maker(fn, bound=True)

        bound = self.__class__(
            fn,
            scope,
            opts=self._opts,
            instance=instance,
            owner=owner,
            name=name,
            key_maker=self._bound_key_maker,
            # values=None if scope is classmethod else self._values,  # FIXME: ?
            values=values,
            value_fn=bound_fn,
        )

        if scope is classmethod and owner is not None:
            setattr(owner, name, bound)
        elif instance is not None:
            instance.__dict__[name] = bound

        return bound

    def __get__(self, instance, owner=None):
        return self._bind(instance, owner)


#


def cached_function(fn=None, **kwargs):  # noqa
    if fn is None:
        return functools.partial(cached_function, **kwargs)

    opts = _CachedFunction.Opts(**kwargs)

    if isinstance(fn, staticmethod):
        return _FreeCachedFunction(fn, opts=opts, value_fn=unwrap_func(fn))

    scope = classmethod if isinstance(fn, classmethod) else None

    return _DescriptorCachedFunction(fn, scope, opts=opts)


def static_init(fn: CallableT) -> CallableT:
    fn = cached_function(fn)
    fn()
    return fn



import abc
import typing as ta

from ... import lang


T = ta.TypeVar('T')
U = ta.TypeVar('U')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class SortedCollection(lang.Abstract, ta.Collection[T]):
    Comparator = ta.Callable[[U, U], int]

    @staticmethod
    def default_comparator(a: T, b: T) -> int:
        """https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons"""

        return (a > b) - (a < b)  # type: ignore

    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def __iter__(self) -> ta.Iterator[T]:
        raise NotImplementedError

    @abc.abstractmethod
    def __contains__(self, value: T) -> bool:  # type: ignore
        raise NotImplementedError

    @abc.abstractmethod
    def add(self, value: T) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def find(self, value: T) -> T | None:
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, value: T) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def iter(self, base: T | None = None) -> ta.Iterable[T]:
        raise NotImplementedError

    @abc.abstractmethod
    def iter_desc(self, base: T | None = None) -> ta.Iterable[T]:
        raise NotImplementedError


class SortedMapping(ta.Mapping[K, V]):
    @abc.abstractmethod
    def items(self) -> ta.Iterator[tuple[K, V]]:  # type: ignore
        raise NotImplementedError

    @abc.abstractmethod
    def items_desc(self) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError

    @abc.abstractmethod
    def items_from(self, key: K) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError

    @abc.abstractmethod
    def items_from_desc(self, key: K) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError


class SortedMutableMapping(ta.MutableMapping[K, V], SortedMapping[K, V]):
    pass


class SortedListDict(SortedMutableMapping[K, V]):
    @staticmethod
    def _item_comparator(a: tuple[K, V], b: tuple[K, V]) -> int:
        return SortedCollection.default_comparator(a[0], b[0])

    def __init__(self, impl: SortedCollection, *args, **kwargs) -> None:
        super().__init__()
        self._impl = impl
        for k, v in lang.yield_dict_init(*args, **kwargs):
            self[k] = v

    @property
    def debug(self) -> ta.Mapping[K, V]:
        return dict(self)

    def __getitem__(self, key: K) -> V:
        item = self._impl.find((key, None))
        if item is None:
            raise KeyError(key)
        return item[1]

    def __setitem__(self, key: K, value: V) -> None:
        self._impl.remove((key, None))
        self._impl.add((key, value))

    def __delitem__(self, key: K) -> None:
        self._impl.remove((key, None))

    def __len__(self) -> int:
        return len(self._impl)

    def __iter__(self) -> ta.Iterator[K]:
        for k, _ in self._impl:
            yield k

    def items(self) -> ta.Iterator[tuple[K, V]]:  # type: ignore
        yield from self._impl.iter()

    def items_desc(self) -> ta.Iterator[tuple[K, V]]:
        yield from self._impl.iter_desc()

    def items_from(self, key: K) -> ta.Iterator[tuple[K, V]]:
        yield from self._impl.iter((key, None))

    def items_from_desc(self, key: K) -> ta.Iterator[tuple[K, V]]:
        yield from self._impl.iter_desc((key, None))



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
This bypasses debuggers attaching to spawned subprocess children that look like python processes. See:

  https://github.com/JetBrains/intellij-community/blob/e9d8f126c286acf9df3ff272f440b305bf2ff585/python/helpers/pydev/_pydev_bundle/pydev_monkey.py
"""
import shlex
import typing as ta

from ..lite.runtime import is_debugger_attached


##


_SUBPROCESS_SHELL_WRAP_EXECS = False


def subprocess_shell_wrap_exec(*cmd: str) -> ta.Tuple[str, ...]:
    return ('sh', '-c', ' '.join(map(shlex.quote, cmd)))


def subprocess_maybe_shell_wrap_exec(*cmd: str) -> ta.Tuple[str, ...]:
    if _SUBPROCESS_SHELL_WRAP_EXECS or is_debugger_attached():
        return subprocess_shell_wrap_exec(*cmd)
    else:
        return cmd



import enum
import typing as ta

from ... import dataclasses as dc


##


class SimpleType(enum.Enum):
    DOUBLE = 'double'
    FLOAT = 'float'
    INT32 = 'int32'
    INT64 = 'int64'
    UINT32 = 'uint32'
    UINT64 = 'uint64'
    SINT32 = 'sint32'
    SINT64 = 'sint64'
    FIXED32 = 'fixed32'
    FIXED64 = 'fixed64'
    SFIXED32 = 'sfixed32'
    SFIXED64 = 'sfixed64'
    BOOL = 'bool'
    STRING = 'string'
    BYTES = 'bytes'


class TypeRef(dc.Frozen):
    name: str


Type: ta.TypeAlias = SimpleType | TypeRef


##


class Field(dc.Frozen):
    name: str
    type: Type
    num: int
    repeated: bool = False


class Message(dc.Frozen):
    name: str
    fields: ta.Sequence[Field]


##


class ProtoFile(dc.Frozen):
    messages: ta.Sequence[Message] | None = None
